/**
 * @file fsd_device.c
 * @brief Implementation of functions for communication with char device, file fsd_device.c
 * @author Michal Podzimek
 *
 * @addtogroup q_mysqlfs_deamon MySQL FS Deamon
 * @{
 * @addtogroup g_mysqlfs_deamon_device Device
 * @{
 * 
 * This file contents declaration of functions which provide connection to 
 * char device and communication with it.
 * 
 */


#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
 
#include "fsd_device.h"
#include "fsd_debug.h"
#include "fsd_defines.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>


/**
 * @brief Open connection to FSC device.
 * 
 * This function opens connection to FSC device. The path to the device is defined in
 * FSC_DEVICE_LOCATION.
 * 
 * @return Positive integer value on success, zero or negative integer else. Value -EBUSY is 
 * returned on attempt to open the char device when already opened. 
 * 
 */
int fsd_chardevice_open() {
	return  open(FSC_DEVICE_LOCATION, O_RDWR);
}


/**
 * @brief Release device.
 * 
 * Close connection to device.
 * 
 * @param fsc_device Open char device file descriptor.
 * @return Close function error code.
 * 
 */
int fsd_chardevice_close(int fsc_device) {
	return close(fsc_device);
}


/**
 * @brief Read message from device.
 * 
 * Read string from char_device. 
 * @param fsc_device Device to be read from.
 * @param buffer Data buffer, storage for read message.
 * @return Number of read chars, error code of read function else.
 * 
 */
int fsd_chardevice_read(int fsc_device, char * buffer) {
	int result;
	if (!fsc_device) {
		return EFSC_DEVICE_NOT_OPEN;
	}

	result = read(fsc_device, buffer, MAX_MESSAGE_LENGTH); 
	switch (result) {
		case -EINVAL:
			PDEBUG("Fatal error: MAX_MESSAGE_LENGTH = (%d) is shorter than availible data from FSC device.\n", MAX_MESSAGE_LENGTH); 
			break;
		case -EFAULT:
			PDEBUG("Fatal error: Data transfer from FSC device in kernel to FSD in user space failed.\n"); 
			break;
	}
	
	return result;	
}


/**
 * @brief Write message to device.
 * 
 * Write buffer content into FSC device. Number of chars in the buffer must be
 * specified.
 * 
 * @param fsc_device Device to be writen to.
 * @param data Buffer containing the message.
 * @param count Number of characters in the buffer.
 * @return Number of written characters.
 * 
 */
int fsd_chardevice_write(int fsc_device, const char* data, int count) {
	int result;
	if (!fsc_device) {
		return EFSC_DEVICE_NOT_OPEN;
	}

	result = write(fsc_device, data, count); 
	switch (result) {
		case -EINVAL:
			PDEBUG("Fatal error: number of chars in the message = (%d) is greater than  FSC device input buffer size.\n", count); 
			break;
		case -EFAULT:
			PDEBUG("Fatal error: Data transfer FSD in user space to FSC device in kernel failed.\n"); 
			break;
	}
	return result;
}


/**
 *
 * @}
 * 
 * @}
 * 
 */
