/**
 * @file fsd_main.c
 * @brief MySQL based File System Deamon
 * @author Michal Podzimek, Michal Ficek, Tomas Pop, MFF UK, 2007
 *
 * @addtogroup q_mysqlfs_deamon MySQL FS Deamon
 * @{
 * @addtogroup g_mysqlfs_deamon_main Main
 * @{
 * 
 * MySQL FS deamon waiting in loop to message. Parse message, execute command and
 * return message
 * Commands are described in file protokol.txt 
 * @see doc/protokol/protokol.txt
 * 
 */
 
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fsd_mysql.h"
#include "fsd_device.h"
#include "fsd_defines.h"
#include "../parser/parser.h"
#include <mysql/mysql.h>
#include <signal.h>


/**
 * The only way how to stop daemon is to send a SIGINT signal to it.
 * The daemon release correctly input and output device and clean mysql connection. 
 */


/** handler for connection to database */
MYSQL *conn;
/** handler for device - char device */
int fsc_device = 0;


/**
 * @brief Deamon finalizing function.
 * 
 * Function closes open char device for communicating between FS Deamon and FS. 
 * It closes DB connection too. It's set as signal handler.
 * 
 * @param sig Signal for defined terminating the deamon. Should be e.g. SIGINT.
 * 
 */
void exit_deamon(int sig) {
	/* release device */
	fsd_chardevice_close(fsc_device);
	/* close connection to database */
	fsd_db_close(conn);

	printf("MySQL filesystem deamon stopped.\n");

	exit(1);
}


/**
 * @brief Wrapper function for writting to char device.
 * 
 * Adds a protokol version to message and write it to the chardevice.
 *  
 * @param fsc_device Device to write to.
 * @param data Data which may be written.
 * @param len Max len of the string.
 * @return Code returned by fsd_chardevice_write.
 * @see fsd_chardevice_write
 * 
 */
int fsd_write_message(int fsc_device, const char *data, int len) {
	char message_out[MAX_MESSAGE_LENGTH]="";
	
	/* make up message */
	strcpy(message_out, FSD_VERSION);
	strcat(message_out, " ");
	memcpy(message_out + strlen(message_out), data, len);
	
	/* compute message len 
	 * len plus length of version plus one for space and one for newline 
	 */
	len = len + strlen(FSD_VERSION) + 1;
	/* trailing zero for easy debugging prints, there is always space to add
	 * this trailing zero, we have sufficient size of output buffer
	 */
	message_out[len] = '\0';	
	PINFO("Result: '%s'\n\n", message_out);
	
	return fsd_chardevice_write(fsc_device, message_out, len); 
}

/**
 * @brief Make string from rd_item structure.
 * 
 * Write name/id/name2/id2/... list of directory items. This list has max size defined by param len.
 * 
 * @param rd_list Pointer to list of directory items.
 * @param len Max len of the string.
 * @param return_text Output buffer.
 * @return 1 list isnt empty.
 * @return 0 list is empty.
 * 
 */
int fsd_write_read_dir(rd_item **rd_list, int len, char *return_text) {
 
 	/* pointer to previus item */
	rd_item *prev;
 	
 	/* length of item, its for computing message len */
 	int item_length;
 	
 	strcpy(return_text, "");
 	
 	/* while we arent on the end of list */
 	while (*rd_list) {
 		
 		/* compute item length = name + id + type + 3 x / */
 		item_length = strlen((*rd_list)->name) + strlen((*rd_list)->id) + strlen((*rd_list)->type) + 3;
 		if (item_length > len) {
 			/* message len with next item is longer then remaining message len. It must stop
 			 * listing and return this message, remaining items will be return in next calling.
 			 */
 			return 1;	
 		}
 		len -= item_length;
 		
 		/* make up item message part */
 		strcat(return_text, (*rd_list)->name);
 		strcat(return_text, "/");
 		strcat(return_text, (*rd_list)->id);
 		strcat(return_text, "/");
 		strcat(return_text, (*rd_list)->type);
 		strcat(return_text, "/");
 		
 		/* turn pointers to next item and release memory */
 		prev = (*rd_list);
 		(*rd_list) = (*rd_list)->next;
 		free(prev);
 		
 	}
 	
 	/* if all list is read return 0 and the function wount be called */
 	return 0;
 }


/**
 * @brief Main function.
 * 
 * At first function open device and connect to database.
 * Main section of this function is infinite loop where new message is delivered,
 * parsed and perform. After performed the outgoing message is sent.
 * At the end of function device and connection to database are closed.
 * 
 * @return 0 if was succesfully stopped.
 * @return Error number on error.
 * 
 */
int main() {

	/* structure for commands */
	struct_command command;

	/* strings for messages - both input and output message */
	char message[MAX_MESSAGE_LENGTH];

	/* buffer for return messages from db functions */
	char ret[MAX_MESSAGE_LENGTH] = "";

	/* pointer to rd_list - internal structure of MySQL */
	rd_item *rd_list = NULL;

	/* size of returned block of data */
	int block_size;
	
	/* buffer for block */
	char block[MAX_BLOCK_LENGTH];
	
 	char offset[MAX_ID_LENGTH]; 
 	char *data_pointer;
 	int pos;
	
	
	/* signal handling */
	signal(SIGINT, exit_deamon);
	
	PINFO("MySQL filesystem deamon started.\n");
	
	/* iniciation, connect to database */
	if (!(conn = fsd_db_connect(server, user, password, database))) {
		return EDB_CONNECT_ERROR;	
	}

	/* open FSC device - input and output device for communication between
	 *  FS in kernel and FSD in user space
	 */
	if ((fsc_device = fsd_chardevice_open()) <= 0) {
		return EFSC_DEVICE_OPEN_ERROR;
	}
	
	/* deamon is prepared start infinite loop */
	while (1) {
		
		/* count of bytes read by char device */
		int result;
		result = fsd_chardevice_read(fsc_device, message);
		
		if (result == -1) { 
			/* when reading is iterrupted by SIGINT signal, 
			 * the deamon should be terminated as well 
			 */
		    exit_deamon(SIGINT);
		}
		
		
		/* parse command */
		switch (parse_message(message, result, &command)) {
			case EPROTOKOL_VERSION_ERROR: { 
				PDEBUG("Command '%s', len=%d parsing failed.\n", message, result);
			
				/* we need to write something to the FSC device, otherewise it will block */
				fsd_write_message(fsc_device, "ER VERSION_FAULT", 16); 
			
				/* next loop in the infinite loop */
				continue;
				break;
		  	}
			case OK: {
				PINFO("Command: '%s' (ver:'%s', cmd:'%s', prm:'%s')\n", message, 
					command.version, command.command, command.param);
				break;
			}
		}


		/* command branching */
		switch (command.command_nr) {
			case TEST_NR: {
				PINFO("Test database connection.\n");
				if (fsd_test(conn) == OK) {
					fsd_write_message(fsc_device, "OK", 2);	
				} 
				else {
					fsd_write_message(fsc_device, "ER TEST_FAILED", 14);
				}
				break;
			}

			case CREATE_NR: {
				PINFO("Create file '%s' in dir '%s'\n", command.data, command.param);
				fsd_create(conn, command.param, command.data, ret);
				strcpy(message, ret);
				fsd_write_message(fsc_device, message, strlen(message));
				break;
			}
		
			case MKDIR_NR: {
				printf("Make dir %s in dir %s\n", command.data, command.param);
				fsd_mkdir(conn, command.param, command.data, ret);
				strcpy(message, ret);
				fsd_write_message(fsc_device, message, strlen(message));
				break;
			}
		
			case CLEAR_FILE_NR: {
				printf("Clear file %s\n", command.param);
				fsd_clear_file(conn, command.param, ret);
				strcpy(message, ret);
				fsd_write_message(fsc_device, message, strlen(message));
				break;
			}
		
	 		case ADD_BLOCK_NR: {
	 			printf("Add block to file %s, size of block: %i\n", command.param, command.data_len);
	 			fsd_add_block(conn, command.param, command.data_len, command.data, ret);
	 			strcpy(message, ret);
	 			fsd_write_message(fsc_device, message, strlen(message));
				break;
	 		}
		
			case REMOVE_FILE_NR: {
				printf("Remove file %s\n", command.param);
				fsd_remove_file(conn, command.param, ret);
				strcpy(message, ret);
				fsd_write_message(fsc_device, message, strlen(message));
				break;
			}
		
			case REMOVE_DIR_NR: {
				printf("Remove dir %s\n", command.param);
				fsd_remove_dir(conn, command.param, ret);
				strcpy(message, ret);
				fsd_write_message(fsc_device, message, strlen(message));
				break;
			}
		
	 		case GET_BLOCK_NR: {
	 			printf("Get block from file %s, offset: %s \n", command.param, command.data);
	 			/* block readed succesfully, copy block to message */
	 			if (fsd_get_block(conn, command.param, command.data, &block_size, block, ret) == OK) {
	 			  strcpy(message, ret);
	 			  memcpy((message + strlen(message)), block, block_size);
				  fsd_write_message(fsc_device, message, strlen(ret) + block_size);
				} 
				/* there was an error return error message (not copy block) */
				else {
				    strcpy(message, ret);
				    fsd_write_message(fsc_device, message, strlen(ret));
				}
				break;
	 		}
		
			case LOOKUP_NR: {
				printf("Lookup for %s in dir %s \n", command.data, command.param);
				fsd_lookup(conn, command.param, command.data, ret);
				strcpy(message, ret);
				fsd_write_message(fsc_device, message, strlen(message));
				break;
			}

			case READ_INODE_NR: {
				printf("Read inode info for inode nr. %s\n", command.param);
				fsd_read_inode(conn, command.param, ret);
				strcpy(message, ret);
				fsd_write_message(fsc_device, message, strlen(message));
				break;
			}

			case FILE_INFO_NR: {
				printf("Info for file with inode number %s \n", command.param);
				fsd_file_info(conn, command.param, ret);
				strcpy(message, ret);
				fsd_write_message(fsc_device, message, strlen(message));
				break;
			}

			case DIR_INFO_NR: {
				printf("Info for dir with inode number %s \n", command.param);
				fsd_dir_info(conn, command.param, ret);
				strcpy(message, ret);
				fsd_write_message(fsc_device, message, strlen(message));
				break;
			}
		
			case READ_DIR_NR: {
	 			printf("Read dir [id = %s]\n", command.param);
	 			
	 			/* if readdir returns items list send the list in loop */
	 			if (fsd_read_dir(conn, command.param, &rd_list, ret) == OK) {
					strcpy(message, ret);
					
					/* prepare message with len witch fit to message len, while there is item in list */
					while (fsd_write_read_dir(&rd_list, 
							MAX_MESSAGE_LENGTH - strlen(message) - strlen(FSD_VERSION) - 1, ret)) {
						strcat(message, ret);
						
						/* send message with part of a list */
						fsd_write_message(fsc_device, message, strlen(message));	
						result = fsd_chardevice_read(fsc_device, message);
						
						/* when reading is iterrupted by SIGINT signal, the deamon should be terminated as well */
						if (result == -1) {
							exit_deamon(SIGINT);
						}
						strcpy(message, "");
					}
					
					/* send the last message */
					strcat(message, ret);
					fsd_write_message(fsc_device, message, strlen(message));
					
				/* there was an error return error message (not send list) */
				} 
				else {
					strcpy(message, ret);
					fsd_write_message(fsc_device, message, strlen(message));
				}	
				break;
 			}
		
	 		case REPLACE_BLOCK_NR: {
				PDEBUG("command.data_len= '%d'\n", command.data_len);
	 			pos = 0;
	 			data_pointer = command.data;
	 			while ((command.data_len > 0) && ((*data_pointer) != ' ')) {
					offset[pos++] = *(data_pointer);
					data_pointer++; 
					(command.data_len)--;
				}
				data_pointer++;
				(command.data_len)--;
				offset[pos]='\0';
	 			printf("Replace block to file %s, size of block: %i, offset: %s\n", command.param, command.data_len, offset);
	 			fsd_replace_block(conn, command.param, offset, command.data_len, command.data, ret);
	 			strcpy(message, ret);
	 			fsd_write_message(fsc_device, message, strlen(message));
				break;
	 		}

			case MOVE_FILE_NR: {
				printf("Move file with inode number %s to dir %s \n", command.param, command.data);
				fsd_move_file(conn, command.param, command.data, ret);
				strcpy(message, ret);
				fsd_write_message(fsc_device, message, strlen(message));
				break;
			}

			case RENAME_FILE_NR: {
				printf("Rename file with inode number %s to name %s \n", command.param, command.data);
				fsd_rename_file(conn, command.param, command.data, ret);
				strcpy(message, ret);
				fsd_write_message(fsc_device, message, strlen(message));
				break;
			}

			case RENAME_DIR_NR: {
				printf("Rename dir with inode number %s to name %s \n", command.param, command.data);
				fsd_rename_dir(conn, command.param, command.data, ret);
				strcpy(message, ret);
				fsd_write_message(fsc_device, message, strlen(message));
				break;
			}

			case TRUNCATE_FILE_NR: {
				printf("Truncate file %s to length: %s\n", command.param, command.data);
	 			fsd_truncate_file(conn, command.param, command.data, ret);
	 			strcpy(message, ret);
	 			fsd_write_message(fsc_device, message, strlen(message));
	 			break;
			}
	
			default: {
				PINFO("Unknown command\n");	
				fsd_write_message(fsc_device, "UNKNOWN_COMMAND", 15);
				break;
			}
		}
	}
	
	/* normal case the code should never come here... */
	exit_deamon(SIGINT);	
    return 0;
}

/**
 *
 * @}
 * 
 * @}
 * 
 */
