/**
 * @file fsd_mysql.c
 * @brief Implementation of mysql commands to db, file fsd_mysql.c
 * @author Michal Podzimek, Michal Ficek, Tomas Pop, MFF UK, 2007
 *
 * @addtogroup q_mysqlfs_deamon MySQL FS Deamon
 * @{
 * @addtogroup g_mysqlfs_deamon_db Database Operations
 * @{
 * 
 * This file contains definitions of functions which established and close connection to 
 * database. And function which execute commands that interacts with database.
 * 
 */
 
 
#include <mysql/mysql.h> 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fsd_debug.h"
#include "fsd_defines.h"


/** macro used to avoid code duplicity in this file used for table unlocking */
#define UNLOCK_TABLES() 									\
	if (mysql_query(conn, "UNLOCK TABLES")) {				\
				fprintf(stderr, "%s\n", mysql_error(conn));	\
				strcpy(return_text, "ER DB_ERROR");			\
				return EDB_ERROR;							\
	}
	
/** macro used to avoid code duplicity in this file used for db query */
#define FSD_MYSQL_QUERY(qtext, args...) 					\
	snprintf(query, sizeof(query), qtext, ## args);			\
	if (mysql_query(conn, query)) {							\
				fprintf(stderr, "%s\n", mysql_error(conn));	\
				strcpy(return_text, "ER DB_ERROR");			\
				return EDB_ERROR;							\
	}

/** macro used to avoid code duplicity in this file used for db query simple */
#define FSD_MYSQL_QUERY_SMPL(qtext, args...) 					\
	snprintf(query, sizeof(query), qtext, ## args);			\
	if (mysql_query(conn, query)) {							\
				fprintf(stderr, "%s\n", mysql_error(conn));	\
				return EDB_ERROR;							\
	}

/**
 * @brief Connect to database.
 * 
 * This function established connection to database described in parameters.
 * 
 * @param server Name of a server.
 * @param user User name to database server.
 * @param password Password to database server.
 * @param database Name of a database.
 * @return MySQL handler.
 * 
 */
MYSQL *fsd_db_connect(char *server, char *user, char* password,
					  char *database)
{
	/* MySQL connection handler*/
	MYSQL *conn;
	conn = mysql_init(NULL);
       
	/* Connect to database */
	if (!mysql_real_connect(conn, server, user, password, database, 0, NULL, 0)) {
		/* if error occurs, write error and return NULL  */
		printf("%s\n", mysql_error(conn));
		return NULL;
	}
	return conn;
}


/**
 * @brief Close connection to database.
 * 
 * This function close connection to database, that handler is passed in parameter.
 * 
 * @param conn Pointer to MySQL handler.
 * 
 */
void fsd_db_close(MYSQL *conn) {
	mysql_close(conn);
}


/**
 * @brief Test database connection.
 *
 * This function tests database connection. Is used when FS is mounted to determine
 * that database is running and is reachable. Sends query for id = 1 from directories.
 * Query return string is ignored.
 * 
 * @param conn MySQL handler.
 * @return OK on success, EDB_ERROR if query failed.
 * 
 */
int fsd_test(MYSQL *conn) {
	MYSQL_RES *res;
	
	/* call query, if error occurs return EDB_ERROR, OK on success */
	if (mysql_query(conn, "SELECT `id` FROM `dirs` WHERE id='1'")) {
		PDEBUG("%s\n", mysql_error(conn));
		return EDB_ERROR;
	} 
	else {
		res = mysql_use_result(conn);
		mysql_free_result(res);
		return OK;
	}
}


/**
 * @brief Determine that directory id exists in db.
 * 
 * This function determine that id exists in table dirs. It try to find out it.
 * If it isnt found function return EDIR_NOT_EXISTS. If it is return value is OK.
 * If some error occurs EDB_ERROR is return value.
 * 
 * @param conn MySQL handler.
 * @param dir Directory id.
 * @return EDIR_NOT_EXISTS - dir not exists.
 * @return OK - dir exists.
 * @return EDB_ERROR - error.
 * 
 */
int fsd_dir_exists(MYSQL *conn, char *dir) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	char query[MAX_QUERY_LENGTH + 1] = "";
	
	/* call query, if error occurs return EDB_ERROR */
	FSD_MYSQL_QUERY_SMPL("SELECT `id` FROM `dirs` WHERE id='%s'", dir);

	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result is empty directory not exists and return EDIR_NOT_EXISTS */
	if ((row = mysql_fetch_row(res)) == NULL) {
		mysql_free_result(res);
		return EDIR_NOT_EXISTS;
	}
	
	/* otherwise directory exists and return OK */
	mysql_free_result(res);
	return OK;
}


/**
 * @brief Determine that file id exists in db.
 * 
 * This function determine that id exists in table files. It try to find out it.
 * If it isnt found function return EFILE_NOT_EXISTS. If it is return value is OK.
 * If some error occurs EDB_ERROR is return value.
 * 
 * @param conn MySQL handler.
 * @param file File id.
 * @return EFILE_NOT_EXISTS - file not exists.
 * @return OK - file exists.
 * @return EDB_ERROR - error.
 * 
 */
int fsd_file_exists(MYSQL *conn, char *file) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	char query[MAX_QUERY_LENGTH + 1] = "";
	
	/* call query, if error occurs return EDB_ERROR */
	FSD_MYSQL_QUERY_SMPL("SELECT `id` FROM `files` WHERE id='%s'", file);
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result is empty file not exists and return EFILE_NOT_EXISTS */
	if ((row = mysql_fetch_row(res)) == NULL) {
		mysql_free_result(res);
		return EFILE_NOT_EXISTS;
	}
	
	/* otherwise file exists and return OK */
	mysql_free_result(res);
	return OK;
}


/**
 * @brief Determine that directory name exists in parent directory.
 * 
 * This function determine that directory name exists in parent directory. It ask db,
 * if there exists directory with passed name and parent dir id.
 * If it isnt found function return EDIR_NAME_NOT_EXISTS. If it is return value is dir id.
 * If some error occurs EDB_ERROR is return value.
 * 
 * @param conn MySQL handler.
 * @param dir Parent directory id.
 * @param dirname Directory name.
 * @return EDIR_NAME_NOT_EXISTS - dir name not exists.
 * @return Dir_id - dir name exists.
 * @return EDB_ERROR - error.
 * 
 */
int fsd_dir_name_exists(MYSQL *conn, char *dir, char *dirname) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	char query[MAX_QUERY_LENGTH + 1] = "";
	char dirname2[MAX_NAME_LENGTH * 2 + 1] = "";
	int dir_id;

	/* escapes special chars for mysql query */
	mysql_real_escape_string(conn, dirname2, dirname, strlen(dirname));
	
	/* call query, if error occurs return EDB_ERROR */
	FSD_MYSQL_QUERY_SMPL("SELECT `id` FROM `dirs` WHERE parent_dir_id='%s' AND name='%s'", dir, dirname2);
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result is empty directory not exists and return EDIR_NAME_NOT_EXISTS */
	if ((row = mysql_fetch_row(res)) == NULL) {
		mysql_free_result(res);
		return EDIR_NAME_NOT_EXISTS;
	}
	
	/* otherwise directory exists and return dir id */
	dir_id = atoi(row[0]);
	mysql_free_result(res);
	return dir_id;
}


/**
 * @brief Determine that file name exists in parent directory.
 * 
 * This function determine that file name exists in directory. It ask db,
 * if there exists file with passed filename and parent dir id.
 * If it isnt found function return EFILE_NAME_NOT_EXISTS. If it is return value is file id.
 * If some error occurs EDB_ERROR is return value.
 * 
 * @param conn MySQL handler.
 * @param dir Parent directory id.
 * @param filename File name.
 * @return EFILE_NAME_NOT_EXISTS - file not exists.
 * @return File_id - file exists.
 * @return EDB_ERROR - error.
 * 
 */
int fsd_file_name_exists(MYSQL *conn, char *dir, char *filename) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	char query[MAX_QUERY_LENGTH + 1] = "";
	char filename2[MAX_NAME_LENGTH * 2 + 1] = "";
	int file_id;

	/* escapes special chars for mysql query */
	mysql_real_escape_string(conn, filename2, filename, strlen(filename));
	
	/* call query, if error occurs return EDB_ERROR */
	FSD_MYSQL_QUERY_SMPL("SELECT `id` FROM `files` WHERE parent_dir_id='%s' AND name='%s'", dir, filename2);
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result is empty file not exists and return EFILE_NAME_NOT_EXISTS */
	if ((row = mysql_fetch_row(res)) == NULL) {
		mysql_free_result(res);
		return EFILE_NAME_NOT_EXISTS;
	}
	
	/* otherwise file exists and return file id */
	file_id = atoi(row[0]);
	mysql_free_result(res);
	return file_id;
}


/**
 * @brief Create new empty file.
 * 
 * This function create new empty file and return its id. If passed parent directory
 * doesnt exists, or passed name exists in directory it returns error. Error value is
 * returned also if there is problem with connection to database. Return_text must be
 * allocated.
 * 
 * @param conn MySQL handler.
 * @param dir Parent directory id.
 * @param filename File name.
 * @param return_text Buffer for return string.
 * @return OK file_id.
 * @return Error constant.
 * 
 */
int fsd_create(MYSQL *conn, char *dir, char *filename, char *return_text) {
	MYSQL_RES *res = NULL;
	MYSQL_ROW row;
	char query[MAX_QUERY_LENGTH + 1] = "";
	char filename2[MAX_NAME_LENGTH * 2 + 1] = "";

	/* escapes special chars for mysql query */
	mysql_real_escape_string(conn, filename2, filename, strlen(filename));
	
	/* lock tables files and dirs */
	FSD_MYSQL_QUERY("LOCK TABLES `dirs` READ, `files` WRITE")
	
	/* determine if directory exists */
	int tst_dir_exists = fsd_dir_exists(conn, dir);
	if (tst_dir_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;
	}
	if (tst_dir_exists == EDIR_NOT_EXISTS) {
		
		UNLOCK_TABLES();
		
		strcpy(return_text, "ER DIR_NOT_EXISTS");	
		return EDIR_NOT_EXISTS;
	}

	/* determine if file name exists in directory */
	int tst_file_name_exists = fsd_file_name_exists(conn, dir, filename);
	if (tst_file_name_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_name_exists != EFILE_NAME_NOT_EXISTS) {
		
		UNLOCK_TABLES()
		
		strcpy(return_text, "ER NAME_EXISTS");
		return ENAME_EXISTS;
	}	

	/* determine if directory name exists in directory */
	int tst_dir_name_exists = fsd_dir_name_exists(conn, dir, filename);
	if (tst_dir_name_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_dir_name_exists != EDIR_NAME_NOT_EXISTS) {
		
		UNLOCK_TABLES()
		
		strcpy(return_text, "ER NAME_EXISTS");
		return ENAME_EXISTS;
	}	
	
	/* all is ok, it can add new file to db and return its id */
	
	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("INSERT INTO `files` (`parent_dir_id`, `name`) VALUES('%s', '%s')", dir, filename2)
	
	/* find out new file id */
	FSD_MYSQL_QUERY("SELECT LAST_INSERT_ID()")
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if error occurs returns error */
	if ((row = mysql_fetch_row(res)) == NULL) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;
	}
	
	/* copy "OK " and new file id to return_text */
	strcpy(return_text, "OK ");
	if (row[0]) {
		strcat(return_text, row[0]);
	}
	
	/* if not id returned it mean error */
	else {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;
	}
	mysql_free_result(res);
	
	UNLOCK_TABLES()
	
	return OK;
}


/**
 * @brief Create new directory.
 * 
 * This function create new directory and return its id. If passed parent directory
 * doesnt exists, or passed name exists in directory it returns error. Error value is
 * returned also if there is problem with connection to database.
 * 
 * @param conn MySQL handler.
 * @param dir Parent directory id.
 * @param dirname Dir name.
 * @param return_text Buffer for return string.
 * @return OK file_id.
 * @return Error constant.
 * 
 */
int fsd_mkdir(MYSQL *conn, char *dir, char *dirname, char *return_text) {
	MYSQL_RES *res = NULL;
	MYSQL_ROW row;
	char query[MAX_QUERY_LENGTH + 1] = "";
	char dirname2[MAX_NAME_LENGTH * 2 + 1] = "";

	mysql_real_escape_string(conn, dirname2, dirname, strlen(dirname));
	
	/* lock tables files and dirs */
	FSD_MYSQL_QUERY("LOCK TABLES `dirs` WRITE, `files` READ")

	/* determine if directory exists */
	int tst_dir_exists = fsd_dir_exists(conn, dir);
	if (tst_dir_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;
	}
	if (tst_dir_exists == EDIR_NOT_EXISTS) {

		UNLOCK_TABLES()
		
		strcpy(return_text, "ER DIR_NOT_EXISTS");
		return EDIR_NOT_EXISTS;
	}

	/* determine if file name exists in directory */
	int tst_file_name_exists = fsd_file_name_exists(conn, dir, dirname);
	if (tst_file_name_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_name_exists != EFILE_NAME_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER NAME_EXISTS");
		return ENAME_EXISTS;
	}	

	/* determine if directory name exists in directory */
	int tst_dir_name_exists = fsd_dir_name_exists(conn, dir, dirname);
	if (tst_dir_name_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_dir_name_exists != EDIR_NAME_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER NAME_EXISTS");
		return ENAME_EXISTS;
	}	
	
	/* all is ok, it can add new directory to db and return its id */

	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("INSERT INTO `dirs` (`parent_dir_id`, `name`) VALUES('%s', '%s')", dir, dirname2)
	
	/* find out new dir id */
	FSD_MYSQL_QUERY("SELECT LAST_INSERT_ID()")
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if error occurs returns error */
	if ((row = mysql_fetch_row(res)) == NULL) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;
	}
	
	/* copy "OK " and new dir id to return_text */
	strcpy(return_text, "OK ");
	if (row[0]) {
		strcat(return_text, row[0]);
	}
	
	/* if not id returned it mean error */
	else {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;
	}
	mysql_free_result(res);
	
	UNLOCK_TABLES()

	return OK;
}


/**
 * @brief Clear file.
 * 
 * This function remove all blocks from file. If passed file doesnt exists
 * it returns error. Error value is returned also if there is problem with
 * connection to database.
 * 
 * @param conn MySQL handler.
 * @param file File id.
 * @param return_text Buffer for return string.
 * @return OK.
 * @return Error constant.
 * 
 */
int fsd_clear_file(MYSQL *conn, char *file, char *return_text) {
	char query[MAX_QUERY_LENGTH + 1] = "";
	
	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `data` WRITE, `files` READ")

	/* determine if file id exists in table files */
	int tst_file_exists = fsd_file_exists(conn, file);
	if (tst_file_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_exists == EFILE_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER FILE_NOT_EXISTS");
		return EFILE_NOT_EXISTS;
	}	
	
	/* all is ok, it can remove all block from file */
	
	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("DELETE FROM `data` WHERE `file_id` = '%s'", file)
	
	strcpy(return_text, "OK");
	
	UNLOCK_TABLES()
	
	return OK;
}


/**
 * @brief Add block to file.
 * 
 * This function add new block to the end of file. If passed file doesnt exists
 * it returns error. Error value is returned also if there is problem with
 * connection to database.
 * 
 * @param conn MySQL handler.
 * @param file File id.
 * @param block_size Size of new block.
 * @param block Block of data.
 * @param return_text Buffer for return string.
 * @return OK.
 * @return Error constant.
 * 
 */
int fsd_add_block(MYSQL *conn, char *file, int block_size, char *block, char *return_text) {
	/* length of block, where all chars may be dubbled like ' is \'*/
	char query[MAX_BLOCK_LENGTH * 2 + 1 + MAX_QUERY_LENGTH + 1] = "";
	int query_len;
	char *query_pointer;
	/* length of block, where all chars may be dubbled like ' is \'*/
	char block2[MAX_BLOCK_LENGTH * 2 + 1] = "";
	
	/* escapes special chars for mysql query */
	block_size = mysql_real_escape_string(conn,  block2, block, block_size);

	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `data` WRITE, `files` READ")

	/* determine if file id exists in table files */
	int tst_file_exists = fsd_file_exists(conn, file);
	if (tst_file_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_exists == EFILE_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER FILE_NOT_EXISTS");
		return EFILE_NOT_EXISTS;
	}	
	
	/* all is ok, it can remove all block from file */
	
	/* make up query and count query length */
	strcpy(query, "INSERT INTO `data` VALUES (NULL, '");
	strcat(query, file);
	strcat(query, "', '");	
	query_pointer = query + strlen(query);
	query_len = strlen(query);
	memcpy(query_pointer, block2, block_size);
	query_len += block_size;
	query_pointer = query + query_len;
	memcpy(query_pointer, "')", 2);
	query_len += 2;
		
	
	/* call query, if error occurs returns error */
	if (mysql_real_query(conn, query, query_len)) {
		fprintf(stderr, "%s\n", mysql_error(conn));
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;
	}
	
	/* copy "OK " to return_text */
	strcpy(return_text, "OK");

	UNLOCK_TABLES()

	return OK;
}


/**
 * @brief Remove file.
 * 
 * This function remove all blocks from file and remove file from db. If passed file doesnt exists
 * it returns error. Error value is returned also if there is problem with
 * connection to database.
 * 
 * @param conn MySQL handler.
 * @param file File id.
 * @param return_text Buffer for return string.
 * @return OK.
 * @return Error constant.
 * 
 */
int fsd_remove_file(MYSQL *conn, char *file, char *return_text) {
	char query[MAX_QUERY_LENGTH + 1] = "";
	
	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `data` WRITE, `files` WRITE")
	
	/* determine if file id exists in table files */
	int tst_file_exists = fsd_file_exists(conn, file);
	if (tst_file_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_exists == EFILE_NOT_EXISTS) {
		
		UNLOCK_TABLES()

		strcpy(return_text, "ER FILE_NOT_EXISTS");
		return EFILE_NOT_EXISTS;
	}	
	
	/* all is ok, it can remove all block from file and remove file */
	
	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("DELETE FROM `data` WHERE `file_id` = '%s'", file)

	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("DELETE FROM `files` WHERE `id` = '%s'", file)
	
	/* copy "OK " to return_text */
	strcpy(return_text, "OK");

	UNLOCK_TABLES()
	
	return OK;
}


/**
 * @brief Remove dir.
 * 
 * This function remove dir from db. An error is returned when dir is not empty. 
 * If passed dir doesnt exists
 * it returns error. Error value is returned also if there is problem with
 * connection to database.
 * 
 * @param conn MySQL handler.
 * @param dir Dir id.
 * @param return_text Buffer for return string.
 * @return OK.
 * @return Error constant.
 * 
 */
int fsd_remove_dir(MYSQL *conn, char *dir, char *return_text) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	char query[MAX_QUERY_LENGTH + 1] = "";
	
	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `dirs` WRITE, `files` READ")

	/* determine if dir id exists in table dirs */
	int tst_dir_exists = fsd_dir_exists(conn, dir);
	if (tst_dir_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_dir_exists == EDIR_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER DIR_NOT_EXISTS");
		return EDIR_NOT_EXISTS;
	}
	
	/* determine if there is file in dir */
	
	/* call query, if error occurs return EDB_ERROR */
	FSD_MYSQL_QUERY("SELECT `id` FROM `files` WHERE parent_dir_id='%s'", dir)
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result isnt empty file exists and return error */
	if ((row = mysql_fetch_row(res)) != NULL) {
		mysql_free_result(res);
		strcpy(return_text, "ER DIR_IS_NOT_EMPTY");
		return EDIR_IS_NOT_EMPTY;
	}
	
	mysql_free_result(res);
	
	/* determine if there is subdir in dir */
	
	/* call query, if error occurs return EDB_ERROR */
	FSD_MYSQL_QUERY("SELECT `id` FROM `dirs` WHERE parent_dir_id='%s'", dir)
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result isnt empty file exists and return error */
	if ((row = mysql_fetch_row(res)) != NULL) {
		mysql_free_result(res);
		strcpy(return_text, "ER DIR_IS_NOT_EMPTY");
		return EDIR_IS_NOT_EMPTY;
	}

	mysql_free_result(res);
				
				
	/* all is ok, it can remove dir from db */

	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("DELETE FROM `dirs` WHERE `id` = '%s'", dir)
	
	/* copy "OK " to return_text */
	strcpy(return_text, "OK");

	UNLOCK_TABLES()

	return OK;
}


/**
 * @brief Get block of file.
 * 
 * This function return block of file defined by offset. If passed file or block
 * doesnt exists it returns error. Error value is returned also if there is problem with
 * connection to database.
 * 
 * @param conn MySQL handler.
 * @param file File id.
 * @param offset Block offset.
 * @param block_size Size of returned block.
 * @param block Returned block of data.
 * @param return_text Buffer for return string.
 * @return OK.
 * @return Error constant.
 * 
 */
int fsd_get_block(MYSQL *conn, char *file, char *offset, int *block_size, char *block, char *return_text) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	char query[MAX_QUERY_LENGTH + 1] = "";
	
	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `data` READ, `files` READ")

	/* determine if file id exists in table files */
	int tst_file_exists = fsd_file_exists(conn, file);
	if (tst_file_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_exists == EFILE_NOT_EXISTS) {

		UNLOCK_TABLES()
		
		strcpy(return_text, "ER FILE_NOT_EXISTS");
		return EFILE_NOT_EXISTS;
	}	
	
	/* all is ok, it can get block from file */	

	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("SELECT data, length(data) FROM `data` WHERE `file_id`='%s' ORDER BY `id` LIMIT %s, 1", file, offset)
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result is empty block not exists and return error */
	if ((row = mysql_fetch_row(res)) == NULL) {
		mysql_free_result(res);
		strcpy(return_text, "ER BLOCK_NOT_EXISTS");
		return EBLOCK_NOT_EXISTS;
	}
	
	/* determine size of block and copy it */
	*block_size = atoi(row[1]);
	memcpy(block, row[0], *block_size);
	
	/* copy "OK " to return_text */
	strcpy(return_text, "OK X ");
	
	/* otherwise file exists and return 1 */
	mysql_free_result(res);

	UNLOCK_TABLES()

	return OK;
}


/**
 * @brief Lookup for file or directory defined by name.
 * 
 * This function determine if there is file or directory in passed parent 
 * directory with passed name. If passed directory doesnt exists
 * it returns error. Error value is returned also if there is problem with
 * connection to database.
 * 
 * @param conn MySQL handler.
 * @param dir Dir id.
 * @param name Lookuped name.
 * @param return_text Buffer for return string.
 * @return NOT_EXISTS - file or directory doesnt exists.
 * @return IS_FILE file_id - there exists file with passed name.
 * @return IS_DIR dir_id - there exists directory with passed name.
 * @return Error constanttant.
 * 
 */
int fsd_lookup(MYSQL *conn, char *dir, char *name, char *return_text) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	char query[MAX_QUERY_LENGTH + 1] = "";
	char name_id_char[12];	

	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `dirs` READ, `files` READ, `data` READ")

	/* determine if dir id exists in table dirs */
	int tst_dir_exists = fsd_dir_exists(conn, dir);
	if (tst_dir_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_dir_exists == EDIR_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER DIR_NOT_EXISTS");
		return EDIR_NOT_EXISTS;
	}

	/* determine if file name exists in table files */
	int tst_file_name_exists = fsd_file_name_exists(conn, dir, name);
	if (tst_file_name_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_name_exists != EFILE_NAME_NOT_EXISTS) {

		strcpy(return_text, "OK IS_FILE ");
		snprintf(name_id_char, sizeof(name_id_char), "%i", tst_file_name_exists);
		strcat(return_text, name_id_char);
		
		/* find file size */
	
		/* call query, if error occurs returns error */
		FSD_MYSQL_QUERY("SELECT SUM(LENGTH(data)) FROM `data` where `file_id`='%s'", name_id_char)
		
		/* parse result */
		res = mysql_use_result(conn);
		
		/* if result is empty block not exists and return error */
		if ((row = mysql_fetch_row(res)) == NULL) {
			mysql_free_result(res);

			UNLOCK_TABLES()

			strcat(return_text, "/0");
			return OK;
		}
		
		/* file doesnt have any block, mysql returns NULL */
	 	if (!row[0]) {
			mysql_free_result(res);

			UNLOCK_TABLES()

			strcat(return_text, "/0");
			return OK;
		}
	
		strcat(return_text, "/");
		strcat(return_text, row[0]);
		mysql_free_result(res);

		UNLOCK_TABLES()

		return IS_FILE;
	}	

	/* determine if directory name exists in table dirs */
	int tst_dir_name_exists = fsd_dir_name_exists(conn, dir, name);
	if (tst_dir_name_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_dir_name_exists != EDIR_NAME_NOT_EXISTS) {

		UNLOCK_TABLES()
		
		strcpy(return_text, "OK IS_DIR ");
		snprintf(name_id_char, sizeof(name_id_char), "%i", tst_dir_name_exists);
		strcat(return_text, name_id_char);

		/* find subdirs count */
	
		/* call query, if error occurs returns error */
		FSD_MYSQL_QUERY("SELECT COUNT(*) FROM `dirs` where parent_dir_id='%s'", name_id_char)
		
		/* parse result */
		res = mysql_use_result(conn);
		
		/* if result is empty block not exists and return error */
		if ((row = mysql_fetch_row(res)) == NULL) {
			mysql_free_result(res);

			UNLOCK_TABLES()
			
			strcpy(return_text, "/0");
			return OK;
		}
		
		strcat(return_text, "/");	
		strcat(return_text, row[0]);
		mysql_free_result(res);

		/* find subfiles count */
	
		/* call query, if error occurs returns error */
		FSD_MYSQL_QUERY("SELECT COUNT(*) FROM `files` where parent_dir_id='%s'", name_id_char)
		
		/* parse result */
		res = mysql_use_result(conn);
		
		/* if result is empty block not exists and return error */
		if ((row = mysql_fetch_row(res)) == NULL) {
			mysql_free_result(res);

			UNLOCK_TABLES()
			
			strcpy(return_text, "/0");
			return OK;
		}
		
		strcat(return_text, "/");	
		strcat(return_text, row[0]);
		mysql_free_result(res);	

		UNLOCK_TABLES()
		
		return IS_DIR;
	}	
	
	/* isnt dir and isnt file, return not exists */
	strcpy(return_text, "ER NOT_EXISTS");

	UNLOCK_TABLES()	
	
	return ENAME_NOT_EXISTS;
}


/**
 * @brief Used for super read inode. Returns information about inode.
 * 
 * This function return information about inode (satisfies super_read)inode_operation)  
 * Error value is returned also if there is problem with
 * connection to database.
 * 
 * @param conn MySQL handler.
 * @param inode_nr Id of inode.
 * @param return_text Buffer for return string.
 * @return EINODE_NOT_EXISTS - inode of file or directory doesnt exists.
 * @return IS_FILE - there exists file with passed name.
 * @return IS_DIR - there exists directory with passed name.
 * @return OK -on command succesfully done, but error ocured on unlocking tables.
 * 
 */
int fsd_read_inode(MYSQL *conn, char *inode_nr, char *return_text) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	char query[MAX_QUERY_LENGTH + 1] = "";

	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `dirs` READ, `files` READ, `data` READ")

	/* determine if dir id exists in table dirs */
	int tst_dir_exists = fsd_dir_exists(conn, inode_nr);
	if (tst_dir_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_dir_exists == OK) {

		UNLOCK_TABLES()
		
		strcpy(return_text, "OK IS_DIR ");
		
		/* find subdirs count */
	
		/* call query, if error occurs returns error */
		FSD_MYSQL_QUERY("SELECT COUNT(*) FROM `dirs` where parent_dir_id='%s'", inode_nr)
		
		/* parse result */
		res = mysql_use_result(conn);
		
		/* if result is empty block not exists and return error */
		if ((row = mysql_fetch_row(res)) == NULL) {
			mysql_free_result(res);

			UNLOCK_TABLES()
			
			strcpy(return_text, "0");
			return OK;
		}
		
		strcat(return_text, row[0]);
		mysql_free_result(res);

		/* find subfiles count */
	
		/* call query, if error occurs returns error */
		FSD_MYSQL_QUERY("SELECT COUNT(*) FROM `files` where parent_dir_id='%s'", inode_nr)
		
		/* parse result */
		res = mysql_use_result(conn);
		
		/* if result is empty block not exists and return error */
		if ((row = mysql_fetch_row(res)) == NULL) {
			mysql_free_result(res);

			UNLOCK_TABLES()
			
			strcpy(return_text, "/0");
			return OK;
		}
		
		strcat(return_text, "/");	
		strcat(return_text, row[0]);
		mysql_free_result(res);		

		UNLOCK_TABLES()
		
		return IS_DIR;
	}	

	/* determine if file name exists in table files */
	int tst_file_exists = fsd_file_exists(conn, inode_nr);
	if (tst_file_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_exists == OK) {

		UNLOCK_TABLES()
		
		strcpy(return_text, "OK IS_FILE ");
				
		/* find file size */
	
		/* call query, if error occurs returns error */
		FSD_MYSQL_QUERY("SELECT SUM(LENGTH(data)) FROM `data` where `file_id`='%s'", inode_nr)
		
		/* parse result */
		res = mysql_use_result(conn);
		
		/* if result is empty block not exists and return error */
		if ((row = mysql_fetch_row(res)) == NULL) {
			mysql_free_result(res);

			UNLOCK_TABLES()
			
			strcat(return_text, "0");
			return OK;
		}
		
		/* file doesnt have any block, mysql returns NULL */
	 	if (!row[0]) {
			mysql_free_result(res);

			UNLOCK_TABLES()
			
			strcat(return_text, "0");
			return OK;
		}
	
		strcat(return_text, row[0]);
		mysql_free_result(res);

		UNLOCK_TABLES()
		
		return IS_FILE;
	}	

	/* isnt dir and isnt file, return not exists */
	strcpy(return_text, "ER NOT_EXISTS");

	UNLOCK_TABLES()
	
	return EINODE_NOT_EXISTS;
}


/**
 * @brief Read files and directories in directory.
 * 
 * This function return all files and directory which have passed dir_id as parent dir id.
 * If passed directory doesnt exists
 * it returns error. Error value is returned also if there is problem with
 * connection to database.
 * 
 * @param conn MySQL handler.
 * @param dir File id.
 * @param rd_list List of dir items.
 * @param return_text Buffer for return string.
 * @return OK and list of dir items in return_text.
 * @return DIR_NOT_EXISTS - directory doesnt exists.
 * @return Error constant.
 * 
 */
int fsd_read_dir(MYSQL *conn, char *dir, rd_item **rd_list, char *return_text) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	char query[MAX_QUERY_LENGTH + 1] = "";
	int item_count = 0;
	char item_count_ch[18];
	
	/* pointer to new rd_item - item of read dir list */
	rd_item *curr = NULL;
	rd_item *prev = NULL;
	
	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `dirs` READ, `files` READ")

	/* determine if dir id exists in table dirs */
	int tst_dir_exists = fsd_dir_exists(conn, dir);
	if (tst_dir_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_dir_exists == EDIR_NOT_EXISTS) {

		UNLOCK_TABLES()
		
		strcpy(return_text, "ER DIR_NOT_EXISTS");
		return EDIR_NOT_EXISTS;
	}
	
	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("SELECT id, name FROM `dirs` WHERE `parent_dir_id`='%s' ORDER BY `name`", dir)
	
	/* parse result */
	res = mysql_use_result(conn);
	(*rd_list) = NULL;
	/* if result is empty block not exists and return error */

	while ((row = mysql_fetch_row(res)) != NULL) {
		item_count++;
		curr = (rd_item *) malloc(sizeof(rd_item));
		strcpy(curr->id, row[0]);
		strcpy(curr->name, row[1]);
		strcpy(curr->type, "d");
		curr->next = NULL;
		if (!(*rd_list)) {
			(*rd_list) = curr;
		}
		if (prev) {
			prev->next = curr;
		}
		prev = curr;
	}
	mysql_free_result(res);
	
	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("SELECT id, name FROM `files` WHERE `parent_dir_id`='%s' ORDER BY `name`", dir)
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result is empty block not exists and return error */
	while ((row = mysql_fetch_row(res)) != NULL) {
		item_count++;
		curr = (rd_item *) malloc(sizeof(rd_item));
		strcpy(curr->id, row[0]);
		strcpy(curr->name, row[1]);
		strcpy(curr->type, "f");
		curr->next = NULL;
		if (!(*rd_list)) {
			*rd_list = curr;
		}
		if (prev) {
			prev->next = curr;
		}
		prev = curr;
	}
	
	mysql_free_result(res);
	
	/* all list is loaded */
	strcpy(return_text, "OK ");
	snprintf(item_count_ch, sizeof(item_count_ch), "%i", item_count);
	strcat(return_text, item_count_ch);
 	strcat(return_text, " ");

	UNLOCK_TABLES()
	
	return OK;
}


/**
 * @brief Return file info - file size.
 * 
 * This function returns information about file.
 * 
 * @param conn MySQL handler.
 * @param file File id.
 * @param return_text Buffer for return string.
 * @return Size of a file.
 * @return Error constant.
 * 
 */
int fsd_file_info(MYSQL *conn, char *file, char *return_text) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	char query[MAX_QUERY_LENGTH + 1] = "";

	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `files` READ, `data` READ")

	/* determine if file name exists in table files */
	int tst_file_exists = fsd_file_exists(conn, file);
	if (tst_file_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_exists == EFILE_NOT_EXISTS) {

		UNLOCK_TABLES()
		
		strcpy(return_text, "ER FILE_NOT_EXISTS");
		return EFILE_NOT_EXISTS;
	}	

	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("SELECT SUM(LENGTH(data)) FROM `data` where `file_id`='%s'", file)
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result is empty block not exists and return error */
	if ((row = mysql_fetch_row(res)) == NULL) {
		mysql_free_result(res);

		UNLOCK_TABLES()
		
		strcpy(return_text, "OK 0");
		return OK;
	}
	
	/* file doesnt have any block, mysql returns NULL */
 	if (!row[0]) {
		mysql_free_result(res);

		UNLOCK_TABLES()
		
		strcpy(return_text, "OK 0");
		return OK;
	}

	/* isnt dir and isnt file, return not exists */
	strcpy(return_text, "OK ");
	strcat(return_text, row[0]);
	mysql_free_result(res);
	
	UNLOCK_TABLES()
	
	return OK;
}


/**
 * @brief Return dir info - count of items.
 * 
 * This function returns information about directory. The only info at this time is
 * count of items in directory.
 * 
 * @param conn MySQL handler.
 * @param dir Dir id.
 * @param return_text Buffer for return string.
 * @return Dir size - count of dir items.
 * @return Error constant.
 * 
 */
int fsd_dir_info(MYSQL *conn, char *dir, char *return_text) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	char query[MAX_QUERY_LENGTH + 1] = "";

	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `dirs` READ, `files` READ")

	/* determine if file name exists in table files */
	int tst_dir_exists = fsd_dir_exists(conn, dir);
	if (tst_dir_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_dir_exists == EDIR_NOT_EXISTS) {

		UNLOCK_TABLES()
		
		strcpy(return_text, "ER DIR_NOT_EXISTS");
		return EDIR_NOT_EXISTS;
	}	

	/* find subdirs count */

	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("SELECT COUNT(*) FROM `dirs` where parent_dir_id='%s'", dir)
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result is empty block not exists and return error */
	if ((row = mysql_fetch_row(res)) == NULL) {
		mysql_free_result(res);

		UNLOCK_TABLES()
		
		strcpy(return_text, "OK 0");
		return OK;
	}
	
	strcpy(return_text, "OK ");	
	strcat(return_text, row[0]);
	mysql_free_result(res);

	/* find subfiles count */

	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("SELECT COUNT(*) FROM `files` where parent_dir_id='%s'", dir)
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result is empty block not exists and return error */
	if ((row = mysql_fetch_row(res)) == NULL) {
		mysql_free_result(res);

		UNLOCK_TABLES()
		
		strcpy(return_text, "/0");
		return OK;
	}
	
	strcat(return_text, "/");	
	strcat(return_text, row[0]);
	mysql_free_result(res);		

	UNLOCK_TABLES()
	
	return OK;
}


/**
 * @brief Replace block of file.
 * 
 * This function replace block of file defined in parametr offset. It updates
 * data in data table.
 * 
 * @param conn MySQL handler.
 * @param file Dir id.
 * @param offset Block offset.
 * @param block_size New block size.
 * @param return_text Buffer for return string.
 * @param block Block data.
 * @return OK.
 * @return Error constant.
 * 
 */
int fsd_replace_block(MYSQL *conn, char *file, char *offset, int block_size, 
					  char *block, char *return_text) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	/* length of block, where all chars may be dubbled like ' is \' + length of query*/
	char query[MAX_BLOCK_LENGTH * 2 + MAX_QUERY_LENGTH + 1] = "";
	int query_len;
	char *query_pointer;
	char block_id[MAX_ID_LENGTH + 1];
	/* length of block, where all chars may be dubbled like ' is \'*/
	char block2[MAX_BLOCK_LENGTH * 2 + 1] = "";

	/* escapes special chars for mysql query */
	block_size = mysql_real_escape_string(conn, block2, block + strlen(offset) + 1, block_size);
	
	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `data` WRITE, `files` READ")

	/* determine if file id exists in table files */
	int tst_file_exists = fsd_file_exists(conn, file);
	if (tst_file_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_exists == EFILE_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER FILE_NOT_EXISTS");
		return EFILE_NOT_EXISTS;
	}	

	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("SELECT id FROM `data` WHERE `file_id`='%s' ORDER BY `id` LIMIT %s, 1", file, offset)

	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result is empty block not exists and return error */
	if ((row = mysql_fetch_row(res)) == NULL) {
		mysql_free_result(res);
		strcpy(return_text, "BLOCK_NOT_EXISTS");
		
		UNLOCK_TABLES()
		
		return EBLOCK_NOT_EXISTS;
	}

	strcpy(block_id, row[0]);
	mysql_free_result(res);
	
	/* all is ok, it can remove all block from file */

	/* make up query and count query len */
	strcpy(query, "UPDATE `data` SET data='");
	query_len = strlen("UPDATE `data` SET data='");
	query_pointer = query + query_len;
	memcpy(query_pointer, block2, block_size);
	query_len += block_size;
	query_pointer = query + query_len;
	strcpy(query_pointer, "' WHERE id='");
	query_len += strlen("' WHERE id='");
	query_pointer = query + query_len; 
	strcpy(query_pointer, block_id);
	query_len += strlen(block_id);
 	query_pointer = query + query_len;
	strcpy(query_pointer, "'");	
	query_len += strlen("'");
	query_pointer = query + query_len;
	*query_pointer = '\0';
	
	/* call query, if error occurs returns error */
	if (mysql_real_query(conn, query, query_len)) {
		fprintf(stderr, "%s\n", mysql_error(conn));
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;
	}
	
	/* copy "OK " to return_text */
	strcpy(return_text, "OK");

	UNLOCK_TABLES()

	return OK;					  	
}


/**
 * @brief Move file.
 * 
 * Move file to directory defined in parameter dir. The file could be directory also.
 * 
 * @param conn MySQL handler.
 * @param file File id.
 * @param dir New parent dir id.
 * @param return_text Buffer for return string.
 * @return OK.
 * @return Error constant.
 * 
 */
int fsd_move_file(MYSQL *conn, char *file, char *dir, char *return_text) {
	char query[MAX_QUERY_LENGTH + 1] = "";
	
	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `files` WRITE, `dirs` WRITE")

	/* determine if file name exists in table files */
	int tst_dir_exists = fsd_dir_exists(conn, dir);
	if (tst_dir_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_dir_exists == EDIR_NOT_EXISTS) {

		UNLOCK_TABLES()
		
		strcpy(return_text, "ER PARENT_NOT_EXISTS");
		return EPARENT_NOT_EXISTS;
	}	

	/* determine if file id exists in table files */
	int tst_file_exists = fsd_file_exists(conn, file);
	if (tst_file_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	
	/* if it is file update value of parent dir id in files table */
	if (tst_file_exists != EFILE_NOT_EXISTS) {

		FSD_MYSQL_QUERY("UPDATE `files` SET `parent_dir_id`= '%s' WHERE `id` = '%s'", dir, file)
	
		strcpy(return_text, "OK");
		UNLOCK_TABLES()
		return OK;

	}	

	/* determine if dir id exists in table files */
	tst_dir_exists = fsd_dir_exists(conn, file);
	if (tst_dir_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	
	/* if it is directory update value of parent dir id in dirs table */
	if (tst_dir_exists != EDIR_NOT_EXISTS) {
		FSD_MYSQL_QUERY("UPDATE `dirs` SET `parent_dir_id`= '%s' WHERE `id` = '%s'", dir, file)
	
		strcpy(return_text, "OK");
		UNLOCK_TABLES()
		return OK;
	
	}
	
	UNLOCK_TABLES()

	strcpy(return_text, "ER FILE_NOT_EXISTS");
	return EFILE_NOT_EXISTS;

}


/**
 * @brief Rename file.
 * 
 * Rename file to name defined in parameter name. It must checked if name is allready in
 * directory.
 * 
 * @param conn MySQL handler.
 * @param file File id.
 * @param name New file name.
 * @param return_text Buffer for return string.
 * @return OK.
 * @return Error constant.
 * 
 */
int fsd_rename_file(MYSQL *conn, char *file, char *name, char *return_text) {
	char query[MAX_QUERY_LENGTH + 1] = "";
	char parent_dir_id[MAX_ID_LENGTH + 1] = "";
	char name2[MAX_NAME_LENGTH + 1] = "";
	MYSQL_RES *res;
	MYSQL_ROW row;

	/* escapes special chars for mysql query */
	mysql_real_escape_string(conn, name2, name, strlen(name));
	
	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `files` WRITE, `dirs` WRITE")

	/* determine if file id exists in table files */
	int tst_file_exists = fsd_file_exists(conn, file);
	if (tst_file_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_exists == EFILE_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER FILE_NOT_EXISTS");
		return EFILE_NOT_EXISTS;
	}	
	
	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("SELECT `parent_dir_id` FROM `files` where `id`='%s'", file)
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result is empty block not exists and return error */
	if ((row = mysql_fetch_row(res)) == NULL) {
		mysql_free_result(res);

		UNLOCK_TABLES()

		strcpy(return_text, "ER FILE_NOT_EXISTS");
		return EFILE_NOT_EXISTS;
	}
		
	strcat(parent_dir_id, row[0]);
	mysql_free_result(res);			
	
	/* determine if file name exists in directory */
	int tst_file_name_exists = fsd_file_name_exists(conn, parent_dir_id, name);
	if (tst_file_name_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_name_exists != EFILE_NAME_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER NAME_EXISTS");
		return ENAME_EXISTS;
	}	

	/* determine if directory name exists in directory */
	int tst_dir_name_exists = fsd_dir_name_exists(conn, parent_dir_id, name);
	if (tst_dir_name_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_dir_name_exists != EDIR_NAME_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER NAME_EXISTS");
		return ENAME_EXISTS;
	}	
	
	/* all is ok, it can remove all block from file */
	
	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("UPDATE `files` SET `name`= '%s' WHERE `id` = '%s'", name, file)
	
	strcpy(return_text, "OK");
	
	UNLOCK_TABLES()
	
	return OK;
}

/**
 * @brief Rename directory.
 * 
 * Rename directory to name defined in parameter name. It must checked if name is allready 
 * in directory.
 * 
 * @param conn MySQL handler.
 * @param dir Dir id.
 * @param name New file name.
 * @param return_text Buffer for return string.
 * @return OK.
 * @return Error constant.
 * 
 */
int fsd_rename_dir(MYSQL *conn, char *dir, char *name, char *return_text) {
	char query[MAX_QUERY_LENGTH + 1] = "";
	char parent_dir_id[MAX_NAME_LENGTH + 1] = "";
	char name2[MAX_NAME_LENGTH + 1] = "";
	MYSQL_RES *res;
	MYSQL_ROW row;
	
	/* escapes special chars for mysql query */
	mysql_real_escape_string(conn, name2, name, strlen(name));
	
	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `files` WRITE, `dirs` WRITE")

	/* determine if file id exists in table files */
	int tst_dir_exists = fsd_dir_exists(conn, dir);
	if (tst_dir_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_dir_exists == EFILE_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER FILE_NOT_EXISTS");
		return EFILE_NOT_EXISTS;
	}	
	
	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("SELECT `parent_dir_id` FROM `dirs` where `id`='%s'", dir)
	
	/* parse result */
	res = mysql_use_result(conn);
	
	/* if result is empty block not exists and return error */
	if ((row = mysql_fetch_row(res)) == NULL) {
		mysql_free_result(res);

		UNLOCK_TABLES()

		strcpy(return_text, "ER FILE_NOT_EXISTS");
		return EFILE_NOT_EXISTS;
	}
		
	strcat(parent_dir_id, row[0]);
	mysql_free_result(res);			
	
	/* determine if file name exists in directory */
	int tst_file_name_exists = fsd_file_name_exists(conn, parent_dir_id, name);
	if (tst_file_name_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_name_exists != EFILE_NAME_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER NAME_EXISTS");
		return ENAME_EXISTS;
	}	

	/* determine if directory name exists in directory */
	int tst_dir_name_exists = fsd_dir_name_exists(conn, parent_dir_id, name);
	if (tst_dir_name_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_dir_name_exists != EDIR_NAME_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER NAME_EXISTS");
		return ENAME_EXISTS;
	}	
	
	/* all is ok, it can remove all block from file */
	
	/* call query, if error occurs returns error */
	FSD_MYSQL_QUERY("UPDATE `dirs` SET `name`= '%s' WHERE `id` = '%s'", name, dir)
	
	strcpy(return_text, "OK");
	
	UNLOCK_TABLES()
	
	return OK;
}

/**
 * @brief Truncate file.
 * 
 * This function truncates file. It cut file or write 0 to the end of file.
 * 
 * @param conn MySQL handler.
 * @param file File id.
 * @param new_file_size Size of new file.
 * @param return_text Buffer for return string.
 * @return OK.
 * @return Error name.
 * 
 */
int fsd_truncate_file(MYSQL *conn, char *file, char *new_file_size, char *return_text) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	/* length of block, where all chars may be dubbled like ' is \'*/
	char query[MAX_BLOCK_LENGTH * 2 + 1 + MAX_QUERY_LENGTH + 1] = "";
	/* length of block, where all chars may be dubbled like ' is \'*/
	char block2[MAX_BLOCK_LENGTH * 2 + 1] = "";
	int query_len;
	char *query_pointer;
	char block_id[MAX_ID_LENGTH + 1] = "";
	long new_file_size_long = atol(new_file_size);
	int fillchar = 0;
	long file_size;
	int block_number, end_block_number, count;

	/* lock tables files and data */
	FSD_MYSQL_QUERY("LOCK TABLES `files` READ, `data` WRITE")

	/* determine if file id exists in table files */
	int tst_file_exists = fsd_file_exists(conn, file);
	if (tst_file_exists == EDB_ERROR) {
		strcpy(return_text, "ER DB_ERROR");
		return EDB_ERROR;	
	}
	if (tst_file_exists == EFILE_NOT_EXISTS) {

		UNLOCK_TABLES()

		strcpy(return_text, "ER FILE_NOT_EXISTS");
		return EFILE_NOT_EXISTS;
	}	

	/* get file length */
	FSD_MYSQL_QUERY("SELECT SUM(LENGTH(data)) FROM `data` where `file_id`='%s'", file)
	res = mysql_use_result(conn);
	
	if ((row = mysql_fetch_row(res)) == NULL) {

		mysql_free_result(res);
	
		UNLOCK_TABLES()

		strcpy(return_text, "OK");
		return OK;
	}

	/* actual file size */
	if (row[0] == NULL) {
		file_size = 0;
	} 
	else {
		file_size = atol(row[0]);
	}

	mysql_free_result(res);
	
	/* shorten the block with new_file_size */	
	if (new_file_size_long <= file_size) {	

		block_number = new_file_size_long / MAX_BLOCK_LENGTH;
		count = new_file_size_long - block_number * MAX_BLOCK_LENGTH;
		
		FSD_MYSQL_QUERY("SELECT data, id FROM `data` WHERE `file_id`='%s' ORDER BY `id` LIMIT %d, 1", file, block_number)
	
		/* parse result */
		res = mysql_use_result(conn);
		if ((row = mysql_fetch_row(res)) == NULL) {
			mysql_free_result(res);
		
			UNLOCK_TABLES()
	
			strcpy(return_text, "OK");
			return OK;
		}
		strcpy(block_id, row[1]);

		count = mysql_real_escape_string(conn, block2, row[0], count);

		/* make up query and count the query length */
		strcpy(query, "UPDATE `data` SET data='");
		query_len = strlen("UPDATE `data` SET data='");
		query_pointer = query + query_len;
		memcpy(query_pointer, block2, count);
		query_len += count;
		query_pointer = query + query_len;
		strcpy(query_pointer, "' WHERE id='");
		query_len += strlen("' WHERE id='");
		query_pointer = query + query_len; 
		strcpy(query_pointer, block_id);
		query_len += strlen(block_id);
		query_pointer = query + query_len;
		strcpy(query_pointer, "'");	
		query_len += strlen("'");
		query_pointer = query + query_len;
		*query_pointer = '\0';

		mysql_free_result(res);
		if (mysql_real_query(conn, query, query_len)) {
			fprintf(stderr, "%s\n", mysql_error(conn));
			strcpy(return_text, "ER DB_ERROR");
			return EDB_ERROR;
		}

		/* delete unused blocks */

		while (1) {
	
			/* call query, if error occurs returns error */
			FSD_MYSQL_QUERY("SELECT `id` FROM `data` where `file_id`='%s' LIMIT %i, 1", file, block_number+1)
			/* parse result */
			res = mysql_use_result(conn);
			
			/* if result is empty block not exists and return error */
			if ((row = mysql_fetch_row(res)) == NULL) {
				mysql_free_result(res);
		
				UNLOCK_TABLES()
		
				strcpy(return_text, "OK");
				return OK;
			}
			
			strcpy(block_id, row[0]);
			mysql_free_result(res);			
		
			
			/* all is ok, it can remove all block from file */
			
			/* call query, if error occurs returns error */
			FSD_MYSQL_QUERY("DELETE FROM `data` WHERE `id` = '%s'", block_id)
			
		}
	} 
	
	/* fill the last block with zeros */
	else {

		block_number = file_size / MAX_BLOCK_LENGTH;
		count = file_size - block_number * MAX_BLOCK_LENGTH;

		if (file_size > 0) {
			FSD_MYSQL_QUERY("SELECT data, id FROM `data` WHERE `file_id`='%s' ORDER BY `id` LIMIT %d, 1", file, block_number)
		
			/* parse result */
			res = mysql_use_result(conn);
			if ((row = mysql_fetch_row(res)) == NULL) {
				mysql_free_result(res);
			
				UNLOCK_TABLES()
		
				strcpy(return_text, "OK");
				return OK;
			}
			strcpy(block_id, row[1]);
	
			memcpy(query, row[0], count);
			memset((char*)query + count, fillchar, MAX_BLOCK_LENGTH - count);

			count = mysql_real_escape_string(conn, block2, query, MAX_BLOCK_LENGTH);

			/* make up query and count the query length */
			strcpy(query, "UPDATE `data` SET data='");
			query_len = strlen("UPDATE `data` SET data='");
			query_pointer = query + query_len;
			memcpy(query_pointer, block2, count);
			query_len += count;
			query_pointer = query + query_len;
			strcpy(query_pointer, "' WHERE id='");
			query_len += strlen("' WHERE id='");
			query_pointer = query + query_len; 
			strcpy(query_pointer, block_id);
			query_len += strlen(block_id);
			query_pointer = query + query_len;
			strcpy(query_pointer, "'");	
			query_len += strlen("'");
			query_pointer = query + query_len;
			*query_pointer = '\0';
	
	
			mysql_free_result(res);
			if (mysql_real_query(conn, query, query_len)) {
				fprintf(stderr, "%s\n", mysql_error(conn));
				strcpy(return_text, "ER DB_ERROR");
				return EDB_ERROR;
			}
		
		}
		/* add blocks with zeros */
		
		memset((char*)query, fillchar, MAX_BLOCK_LENGTH);
		count = mysql_real_escape_string(conn, block2, query, MAX_BLOCK_LENGTH);

		/* make up query and count the query length */
		strcpy(query, "INSERT INTO `data` VALUES (NULL, '");
		strcat(query, file);
		strcat(query, "', '");	
		query_pointer = query + strlen(query);
		query_len = strlen(query);
		memcpy(query_pointer, block2, count);
		query_len += count;
		query_pointer = query + query_len;
		memcpy(query_pointer, "')", strlen("')"));
		query_len += strlen("')");

		end_block_number = new_file_size_long / MAX_BLOCK_LENGTH;
		count = new_file_size_long - end_block_number * MAX_BLOCK_LENGTH;
		
		while (block_number++ < end_block_number) {
			if (mysql_real_query(conn, query, query_len)) {
				fprintf(stderr, "%s\n", mysql_error(conn));
				strcpy(return_text, "ER DB_ERROR");
				return EDB_ERROR;
			}
		}

		memset((char*)query, fillchar, count);
		count = mysql_real_escape_string(conn, block2, query, count);

		/* make up query and count the query length */
		strcpy(query, "INSERT INTO `data` VALUES (NULL, '");
		strcat(query, file);
		strcat(query, "', '");	
		query_pointer = query + strlen(query);
		query_len = strlen(query);
		memcpy(query_pointer, block2, count);
		query_len += count;
		query_pointer = query + query_len;
		memcpy(query_pointer, "')", strlen("')"));
		query_len += strlen("')");
		query_pointer = query + query_len;
		*query_pointer = '\0';
		if (mysql_real_query(conn, query, query_len)) {
			fprintf(stderr, "%s\n", mysql_error(conn));
			strcpy(return_text, "ER DB_ERROR");
			return EDB_ERROR;
		}

	}

	UNLOCK_TABLES()

	strcpy(return_text, "OK");
	return OK;
}


/**
 *
 * @}
 * 
 * @}
 * 
 */
