/**
 * @file fs_dir.c
 * @brief Defines functions for kernel part of filesystem to manipulate with directories
 * @author Michal Podzimek, Michal Ficek, Tomas Pop, MFF UK, 2007
 *
 * 
 * This file implements common operations on directories like mkdir and rmdir
 * 
 * @addtogroup q_mysqlfs_kernel MySQL FS Kernel Module
 * @{
 * @addtogroup g_mysqlfs_kernel_dir_c Directory operations
 * @{
 * 
 * 
 */


#ifndef __KERNEL__
#	define __KERNEL__
#endif


#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/statfs.h>
#include <linux/mm.h>
#include <asm/uaccess.h>
#include <asm/errno.h>
#include <linux/buffer_head.h>
#include <linux/pagemap.h> /* unlock_page */
#include <linux/proc_fs.h>
#include <asm/atomic.h>
#include <linux/string.h>


#include "fsc_device.h"
#include "rkfs.h"
#include "parser.h"


/** dir operations structure */
struct file_operations rkfs_dops = {
	read: generic_read_dir,
	readdir: rkfs_file_readdir,
	fsync: simple_sync_file
};


/**
 * @brief Function for make new directory
 * 
 * Create new directory in parent directory specified by dir parametr
 * 
 * @param dir Directory, where new dir should be created.
 * @param dentry Dentry representing name of directory, which should be created.
 * @param mode Mode of new directory.
 * @return 0 OK, or -EMLINK on DB error or PTR_ERR(inode) on inode creation error.
 * 
 */
int rkfs_mkdir(struct inode * dir, struct dentry *dentry, int mode) {
	struct inode * inode;
	struct super_block *sb = dir->i_sb;
	int err = -EMLINK;
	char query_string[MAX_QUERY_LENGTH + 1];

	snprintf(query_string, MAX_QUERY_LENGTH + 1, "MYSQL1.1 MD %lu %s", dir->i_ino, dentry->d_name.name);

	if (query(query_string, strlen(query_string)) != RKFS_DEF_QUERY_OK) {		
		return err;
	} 
	else {
		if (ret_command.command_nr == CMD_ER_NR) {
			return err;
		}
	}

	inode_inc_link_count(dir);
	
	/* allocate an inode */
	inode = iget(sb, rkfs_atoi(ret_command.param));
	
	err = PTR_ERR(inode);
	if (IS_ERR(inode)) {
		inode_dec_link_count(dir);
		return err;
	}
	/* size of the new directory is 0 - no files inside */
	inode->i_size = 0; 
	
	/* create as dir, with wanted mode */
	inode->i_mode = S_IFDIR | mode; 
	
	/* set the directory ops */
	inode->i_fop = &rkfs_dops;
	
	/* set the inode ops */ 
	inode->i_op = &rkfs_iops;
	
	/* number of hardlinks for the new directory is 2 - one for . and one for .. directory */ 
	inode->i_nlink =2; 

	d_instantiate(dentry, inode);
	
	/* parent directory size increases by one for this new directory */ 
	dir->i_size++;
	return 0; /* ok */
}


/**
 * @brief Function for remove directory
 * 
 * Remove directory specified in dentry
 *  
 * @param dir Dir from which should be dir removed.
 * @param dentry Dentry representing name of directory, which should be removed.
 * @return OK on succes or -ENOTEMPTY when directory not empty, -EFAULT on other errors.
 * 
 */
int rkfs_rmdir(struct inode * dir, struct dentry *dentry) {
	/* get pointer to a coresponding inode */
	struct inode * inode = dentry->d_inode;
	int err = -ENOTEMPTY;
	char query_string[MAX_QUERY_LENGTH + 1];

	FSPDEBUG("rkfs_rmdir for file '%s'\n", dentry->d_name.name);

	if (inode->i_size == 0) {
		snprintf(query_string, MAX_QUERY_LENGTH + 1, "MYSQL1.1 RD %lu", inode->i_ino);
		if (query(query_string, strlen(query_string)) != RKFS_DEF_QUERY_OK) {
			return -EFAULT;
		} 
		else {
			if (ret_command.command_nr == CMD_ER_NR) {
				return -EFAULT;
			}
		}
		inode->i_size = 0;
		inode_dec_link_count(inode);
		inode_dec_link_count(dir);
		
		/* parent directory size decreases by one for this removed directory */ 
		dir->i_size--;
		return 0;
	} 
	else {
		/* directory was not empty */
		return err;
	} 
}


/**
 *
 * @}
 * 
 * @}
 * 
 */
