/**
 * @file fsc_device.c 
 * @brief FS Char Device settings, implementation
 * 
 * @author Michal Podzimek, Michal Ficek, Tomas Pop, MFF UK, 2007
 * 
 * This file implements special char device used for comunication between 
 * kernel part of a filesystem and userspace deamon. Be carefull if using it
 * this char device has very special behaviour, in some cases it behaves more like 
 * block device
 * Especialy We mean this two features
 * <ul>
 * <li>All the data in buffer are read at once and the real lenght of data is returned</li>
 * <li>After reading all the data never 0 is returned like in clasical chardevices</li>
 * </ul>
 * @addtogroup q_mysqlfs_kernel MySQL FS Kernel Module
 * @{
 * @addtogroup g_mysqlfs_kernel_chardev MySQL FS Char Device
 * @{
 *
 */


#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/statfs.h>
#include <linux/mm.h>
#include <asm/uaccess.h>
#include <asm/errno.h>
#include <linux/buffer_head.h>
#include <linux/pagemap.h> 	/* unlock_page */
#include <linux/proc_fs.h>
#include <asm/atomic.h>


#include "fsc_debug.h"
#include "fsc_device.h"


/** Output buffer stores data to be read from char device by the deamon. */
struct buffer fsc_output_buffer;


/** Input buffer stores data to be written to char device by the deamon. */
struct buffer fsc_input_buffer;	


/** Atomic variable for determinig usage of fsc device by exactly one user at the moment. */
atomic_t fsc_available = ATOMIC_INIT(1);


/** Semaphore for mutual exclusive access to fsc device buffers. */
struct semaphore fsc_sem; 


/** Structure for mutex */
struct mutex query_mutex;


/** FSC Device module major number. */
int fsc_major = 0;


/** Expose fsc device major number. */
module_param(fsc_major, int, S_IRUSR);


/* char device functions declarations */
ssize_t fsc_read(struct file *, char *, size_t, loff_t *);
ssize_t fsc_write(struct file *, const char *, size_t, loff_t *);
int fsc_release(struct inode *, struct file *);
int fsc_open(struct inode *, struct file *);


/** 
 * This kernel structure contains informations about operations supported by fsc device.
 */
struct file_operations fsc_fops = {
  .read = fsc_read,        /**< Reading from char device. */
  .write = fsc_write,      /**< Writing to char device. */
  .open = fsc_open,        /**< Open the char device. */
  .release = fsc_release   /**< Release the char device. */
};


/**
 * @brief FS Character Device initialization
 * 
 * Initializes wait queues for both input and output buffer, declares buffer 
 * contents as empty and initializes semaphore for mutual exclusive access to 
 * char device internals.
 * 
 */
void fsc_init(void) {
    fsc_output_buffer.length = 0;
    fsc_input_buffer.length = 0;
    init_MUTEX(&fsc_sem);
    mutex_init(&query_mutex);
    init_waitqueue_head(&fsc_output_buffer.deamon_queue);
    init_waitqueue_head(&fsc_output_buffer.kernel_queue);
    init_waitqueue_head(&fsc_input_buffer.deamon_queue);
    init_waitqueue_head(&fsc_input_buffer.kernel_queue);
}


/**
 * @brief FS Character Device fini
 * 
 * Destroy mutex.
 * 
 */
void fsc_fini(void) {
    mutex_destroy(&query_mutex);
}


/**
 * @brief Prints information about the fsc device
 * 
 * Not necessary for running, but useful for checking the char device buffer content. 
 * Only way to see what is stored inside the buffers withnout reading or writing into them.
 * Params usually not needed, call "cat /proc/fsc_dev".
 * 
 * @param buf Buffer will contain char device info.
 * @param start Ignored.
 * @param offset Ignored.
 * @param count Ignored.
 * @param eof End Of File parameter.
 * @param data Ignored.
 * @return Number of chars copied to buffer buf.
 * 
 */
int fsc_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data) {
		
  *eof=1; /* No further reading after this info print */	
  return sprintf(buf, "FSC input buffer: %s\n"
		     "FSC input buffer length declared: %d\n"
		     "FSC output buffer: %s\n"
		     "FSC output buffer length declared: %d\n",
		      fsc_input_buffer.data, fsc_input_buffer.length, 
		      fsc_output_buffer.data, fsc_output_buffer.length);
}

 
/**
 * @brief Open the char device.
 * 
 * Char device is open exclusively, only one open instance in the time.
 * 
 * @param inode Ignored.
 * @param filp Ignored.
 * @return Zero on success, -EBUSY if device already open.
 * 
 */
int fsc_open(struct inode *inode, struct file *filp) {
	PDEBUG("fsc_open(%p, %p), major=%d minor=%d\n", inode, filp, 
			MAJOR(inode->i_rdev), MINOR(inode->i_rdev));
	if (! atomic_dec_and_test (&fsc_available)) {
		atomic_inc(&fsc_available);
		return -EBUSY;
	}
	
	try_module_get(THIS_MODULE);
	return 0;
}


/**
 * @brief Release (close) the char device.
 * 
 * @param inode Ignored.
 * @param filp Ignored.
 * @return Zero on success.
 * 
 */
int fsc_release(struct inode *inode, struct file *filp) {
	atomic_inc(&fsc_available);
	PDEBUG("fsc_release(%p, %p), major=%d minor=%d\n", inode, filp, 
			MAJOR(inode->i_rdev), MINOR(inode->i_rdev));
	module_put(THIS_MODULE);
	return 0;
}


/**
 * @brief Wait for empty output buffer.
 * 
 * Output buffer is used for storing FS commands and data until FS Deamon reads it. 
 * When new attempt from kernel to write into ouput buffer is processed and old data are present,
 * this function blocks current running kernel thread using wait queue. The queue is
 * defined in the output buffer sturcture, kernel always sleeps in the kernel_queue.
 * Caller of this function has to hold the semaphore. After return from waiting 
 * with success, the semaphore is set down again.
 * 
 * @return Zero on success (output_buffer.length == 0), -ERESTARTSYS if waitinig was interrupted
 * by the signal, or char device after waiging is used  (semaphore is already down)
 * 
 */
int fsc_waitforempty_output(void) {
	while (fsc_output_buffer.length != 0) {
		up(&fsc_sem);
		PDEBUG("fsc_waitforempty_output (len=%d) - going to sleep\n", fsc_output_buffer.length);
		if (wait_event_interruptible(fsc_output_buffer.kernel_queue, fsc_output_buffer.length == 0)) {
			/* signal: tell the fs layer to handle it */
			return -ERESTARTSYS; 
		}
		
		if (signal_pending(current)) {
			/* signal: tell the fs layer to handle it */
			return -ERESTARTSYS; 
		}
		
		if (down_interruptible(&fsc_sem)) {
			/* semaphore: char device in use */
			return -ERESTARTSYS; 
		}
		
	}
	return 0;
}


/**
 * @brief Wait for full input buffer.
 * 
 * Input buffer is used for storing response from database or command response written by FS Deamon.
 * When new attempt from kernel to read from input buffer is processed and no data are present,
 * this function blocks current running kernel thread using wait queue. The queue is
 * defined in the input buffer structure, kernel alwazs sleeps in the kernel_queue.
 * Calle of this function has to hold the semaphore. After return from waitin with success,
 * the semaphore is set down again. 
 * 
 * @return Zero on success (input_buffer.lenght != 0), -ERESTARTSYS if waiting was interrupted
 * by the signal, or char device after waiging is used (semaphore already down).
 * 
 */
int fsc_waitforfull_input(void) {
	while (fsc_input_buffer.length == 0) {
		up(&fsc_sem);
		PDEBUG("fsc_waitforfull_input (len=%d) - going to sleep\n", fsc_input_buffer.length);
		if (wait_event_interruptible(fsc_input_buffer.kernel_queue, fsc_input_buffer.length != 0)) {
			/* signal: tell the fs layer to handle it */
			return -ERESTARTSYS; 
		}
		
		if (signal_pending(current)) {
			/* signal: tell the fs layer to handle it */
			return -ERESTARTSYS; 
		}
		
		if (down_interruptible(&fsc_sem)) {
			/* semaphore: char device in use */
			return -ERESTARTSYS; 
		}
		
	}
	return 0;
}


/* forward declaration */
int __fsc_query__(const char*, const int, char*, int);


/** Wraper for __fsc_query__() */
int fsc_query(const char * out_buf, const int out_count, char* in_buf, int in_count) {
  static int result;
  mutex_lock(&query_mutex);
  result = __fsc_query__(out_buf, out_count, in_buf, in_count);
  mutex_unlock(&query_mutex);
  return result;
}

/**
 * @brief File System Char Device communication channel.
 * 
 * This function writes FS query into output buffer, waits until FS Deamon read it,
 * waits until FS Deamon writes the answer via write function. 
 * This function can will block in few situations:
 * 1) the FS sends query but output_buffer is still full (hasn't been read)
 * 2) the FS Char Dev waits until input_buffer contains some data (answer)
 * fsc_query function is always called by kernel filesystem, the return value should be 
 * carefully checked and the query should be repeated if error occurs.
 * 
 * @param out_buf Command or data from FS to FS Deamon. Note: 
 * @param out_count Length of out_buf content (number of chars).
 * @param in_buf Buffer for storing the answer or returning data. 
 * @param in_count Maximum length of the buffer, max number of char that could be written into. 
 * @return FS Deamon response length, number of characters written into input_buffer via write function. 
 *         -ERESTARTSYS if the char device is currently used or waiting was interrupted by signal
 *         -EINVAL if query length is greater then char device output buffer size
 * 
 */
int __fsc_query__(const char * out_buf, const int out_count, char* in_buf, int in_count) {
	int i;
	int result;
	int response_len;
	PDEBUG("fsc_query(%s, %d, %s, %d)\n", out_buf, out_count, in_buf, in_count);
	if (down_interruptible(&fsc_sem)) { /* lock the char device internals */
		PDEBUG("fsc_query returning error on semaphore: -ERESTARTSYS\n");
		return -ERESTARTSYS;
	}
	
	result = fsc_waitforempty_output();
	
	/* if error occurs, return it */
	if (result) {
		PDEBUG("fsc_query returning error from waitforempty_output - ERESTARTSYS: %d\n", result);
		/* up called by waitforempty_output() */
		return result; 
	}
	
	if (out_count >= MAX_BUFFER_LENGTH) {	/* just for case */
		up(&fsc_sem);
		return -EINVAL;
	}
	
	/* copy the out_buffer to output_buffer */
	for (i=0; i<out_count; i++) {  
		fsc_output_buffer.data[i] = out_buf[i];
	}
	
	fsc_output_buffer.data[out_count] = '\0';
	fsc_output_buffer.length = out_count; 
	up(&fsc_sem);
	PDEBUG("output buffer filled, waking up sleeping deamon reader\n");
	
	/* the buffer is full, wake up sleaping reader - FS Deamon */
	wake_up_interruptible(&fsc_output_buffer.deamon_queue); 

	/* hold the semaphore lock */
	if (down_interruptible(&fsc_sem)) { 
		PDEBUG("fsc_query returning error on semaphore: -ERESTARTSYS\n");
		return -ERESTARTSYS;
	}
	
	result = fsc_waitforfull_input();
	if (result) {
		PDEBUG("fsc_query returning error from waitforempty_output - ERESTARTSYS: %d\n", result);
		/* up called by waitforempty_output() */
		return result; 
	}
	
	/* copy the input buffer to in_buf */
	for (i = 0; i<fsc_input_buffer.length && i<in_count; i++) {
		/* the leading zero isn't copied, for writing to log add it later */ 
		in_buf[i] = fsc_input_buffer.data[i];	
	}
	
	/* storing number of recieved chars in the response */
	response_len = fsc_input_buffer.length; 
        in_buf[response_len] = '\0';
        /* declare input buffer as empty, for the deamon can fill it again */
        fsc_input_buffer.length = 0; 
	up(&fsc_sem); 
	PDEBUG("input buffer read, waking up sleeping deamon writer\n");
	
	/* the buffer is empty, the FS Deamon writer can be woken up again. */
	wake_up_interruptible(&fsc_input_buffer.deamon_queue); 
	return response_len;
}


/**
 * @brief Char Device write function.
 * 
 * Allways called from user space by FS Deamon. The deamon writes command answers and data from database.
 * On full input_buffer (non-zero declared length) the caller will be blocked - added into wait queue.
 * The caller will be woken up after reading full content of input buffer from kernel FS. 
 * This function don't support writing larger ammount of data then input buffer size. 
 * 
 * @param filp Ignored.
 * @param buf Buffer with written data.
 * @param count Number of chars in the buf buffer.
 * @param f_pos Ignored.
 * @return Number of written characters. - ERESTARTSYS if the char device internals is currently used 
 *         or waiting was interrupted by signal. -EINVAL if count of chars for writting is greater
 *         then char device input buffer size. -EFAULT if data transfer from user space to kernel space failed.
 *
 */
ssize_t fsc_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos) {
  PDEBUG("fsc_write(%p, %p,  %d, %lld)\n", filp, buf, (int) count, *f_pos);
  if (count > MAX_BUFFER_LENGTH) {
  	return -EINVAL;
  }
  
  /* lock the char device internals */
  if (down_interruptible(&fsc_sem)) { 
    return -ERESTARTSYS;
  }
  
  /* wait for empty input buffer */
  while (fsc_input_buffer.length != 0) { 
    up(&fsc_sem); 
    PDEBUG("attemt for writing, buffer full, going to sleep\n");
    
    /* the user space deamon waits allways in deamon_queue, kernel has got it's own */
    if (wait_event_interruptible(fsc_input_buffer.deamon_queue, fsc_input_buffer.length == 0)) {
      /* signal: tell the fs layer to handle it */	
      return -ERESTARTSYS; 
    }
    
    if (down_interruptible(&fsc_sem)) {
      /* semaphore: char device internals in use */
      return -ERESTARTSYS; 
    }
    
  }
  
  /* buffer is clear, let's write something
   * copy +1 byte for trailing zero - added for debugging printouts 
   */
  if (copy_from_user(fsc_input_buffer.data, buf, count + 1)) {  
	up(&fsc_sem); 
	/* on error release the lock first */
  	return -EFAULT;
  }
  
  /* save number of written characters */
  fsc_input_buffer.length = count; 
  up(&fsc_sem);
  PDEBUG("buffer full, waking up sleeping kernel reader\n");
  
  /* the kernel can now read the content of input_buffer */
  wake_up_interruptible(&fsc_input_buffer.kernel_queue); 
  fsc_input_buffer.data[fsc_input_buffer.length] = '\0';
  PDEBUG("db response stored in fsc_input_buffer: '%s'\n", fsc_input_buffer.data);
  
  /* allways all chars written */ 
  return count; 
}


/**
 * @brief Char Device read function.
 * Allways called from user space by FS Deamon. The deamon reads content of output_buffer 
 * and sends it as commands and data to database.
 * On empty output buffer (zero declared length) the caller will be blocked - added into wait queue.
 * The caller will be woken up after writing any data into output buffer from kernel FS. 
 * This function don't support reading larger ammount of data then output buffer size. 
 * Greater number of char to be read
 * can be specified, but the maximum ouput buffer length is returned.
 * If count is less than output_buffer length, count characters will be read, but the kernel 
 * FS remains blocked until whole output_buffer 
 * is read.
 * 
 * @param filp Ignored.
 * @param buf Buffer for storing output_buffer data.
 * @param count Number of chars to be read.
 * @param f_pos Ignored.
 * @return Number of read characters. - ERESTARTSYS if the char device internals is currently used 
 *         or waiting was interrupted by signal. -EFAULT if data transfer from kernel to 
 * 		   user space failed. -EINVAL if user gives smaller buffer than availible 
 *
 */
ssize_t fsc_read(struct file *filp, char *buf, size_t count, loff_t *f_pos) {
  int tmp_count;

  PDEBUG("fsc_read(%p, %p, %d, %lld)\n", filp, buf, (int) count, *f_pos);
  if (down_interruptible(&fsc_sem)) {
	  return -ERESTARTSYS;
  }
  
  /* wait for non-emtpy output buffer */
  while (fsc_output_buffer.length == 0) { 
    up(&fsc_sem);
    PDEBUG("attemt for reading, buffer empty, going to sleep\n");
    
    /* the user space deamon waits allways in deamon_queue, kernel has got it's own */
    if (wait_event_interruptible(fsc_output_buffer.deamon_queue, fsc_output_buffer.length != 0)) {
      /* signal: tell the fs layer to handle it */
      return -ERESTARTSYS; 
    }
    
    if (down_interruptible(&fsc_sem)) {
      /* otherewise loop, but first reacquire the lock */
      return -ERESTARTSYS;
    }
    
  }
  
  /* User wants less data then present. It generates error, 
   * becouse fsc dont support countinuous reads 
   */
  if (count < fsc_output_buffer.length) {	
  	return -EINVAL;
  }
  
  tmp_count = fsc_output_buffer.length;
  fsc_output_buffer.data[fsc_output_buffer.length]='\0';
  
  /* copy +1 byte for trailing zero - added for debugging printouts */
  if (copy_to_user(buf, fsc_output_buffer.data, fsc_output_buffer.length + 1)) { 
	up(&fsc_sem); 
	/* on error release the lock first */
  	return -EFAULT;
  }
  
  fsc_output_buffer.length = 0;	
  up(&fsc_sem);
  PDEBUG("buffer empty, waking up sleeping writers\n");
  wake_up_interruptible(&fsc_output_buffer.kernel_queue); 
  
  /* the kernel can now fill the output buffer with commands and data again */
  PDEBUG("db request fetched from fsc_output_buffer: %s\n", fsc_output_buffer.data); 
  return tmp_count;
} 


/**
 *
 * @}
 * 
 * @}
 * 
 */
