/**
 * @file rkfs.h
 * @brief Declares functions and constants for kernel part of filesystem
 * @author Michal Podzimek, Michal Ficek, Tomas Pop, MFF UK, 2007
 *
 * @addtogroup q_mysqlfs_kernel MySQL FS Kernel Module
 * @{
 * @addtogroup g_mysqlfs_kernel_rkfs_h Header file
 * @{
 * 
 */


#ifndef FS_H_
#define FS_H_

#include "fsc_device.h"
#include "../parser/parser.h"


/** @name File system constants
 * 
 * @{
 */
#define RKFS_MAGIC 		0xabcd
#define RKFS_NAME 		"rkfs"
#define RKFS_BLOCKSIZE  	4096
#define RKFS_BLOCKSIZE_BITS 	10
/** base for directory ids 
 * The inode id's space is divided into two parts. The lower one from 1 to RKFS_DIRECTORY_BASE is
 * used for directory inode numbers, the upper one from RKFS_DIRECTORY_BASE + 1 to max(unsigned long 
 * int) for file inode numbers. */
#define RKFS_DIRECTORY_BASE	10000	
#define RKFS_ROOT_INODE		RKFS_DIRECTORY_BASE + 1
/** @} */


/** query return constant OK */
#define RKFS_DEF_QUERY_OK	 	0
/** query return constant ERROR */
#define RKFS_DEF_QUERY_ERROR	-1


/** RKFS file inode number - delete it */
#define FILE_INODE_NUMBER 2

/** length of output buffer, its same as buffer length */
#define OUT_BUF_LENGTH MAX_BUFFER_LENGTH


/** If defined, debugging prints will be visible. */
#define FS_DEBUG


/** @name Macros to help debugging
 * 
 * @{
 */
#undef FSPDEBUG             /* undef it, just in case */
#ifdef FS_DEBUG

	/* This one if debugging is on, and kernel space */
#	ifdef __KERNEL__

#		define FSPDEBUG(fmt, args...) printk( KERN_DEBUG "RKFS: " fmt, ## args)
#		define FSPINFO(fmt, args...) printk( KERN_INFO "RKFS: " fmt, ## args)
#		define FSPRETCMD() printk(KERN_DEBUG "RKFS: ret_command=(ver:'%s', cmd:'%s', prm:'%s', dta:'%s')\n", ret_command.version, ret_command.command, ret_command.param, ret_command.data)
    
	/* This one for user space */
#	else

#		define FSPDEBUG(fmt, args...) fprintf(stderr, fmt, ## args)
#		define FSPINFO(fmt, args...) fprintf(stdout, fmt, ## args)
#		define FSPRETCMD() fprintf (stderr, "RKFS: ret_command=(ver:'%s', cmd:'%s', prm:'%s', dta:'%s')\n",  ret_command.version, ret_command.command, ret_command.param, ret_command.data)

#	endif
#else

#	define FSPDEBUG(fmt, args...) /* not debugging: nothing */
#	define FSPINFO(fmt, args...) 
#	define FSPRETCMD()

#endif
/** @} */

/** @name declaration of functions defines in fs_inode.c
 *  @see fs_inode.c
 * 
 * @{
 */
extern int rkfs_get_sb (struct file_system_type *fs_type, int flags, const char *devname, 
						void *data, struct vfsmount *mnt);
extern void rkfs_kill_sb(struct super_block *);
extern void rkfs_super_read_inode(struct inode *inode);
extern int rkfs_super_write_inode(struct inode *inode, int sync);
extern void rkfs_super_delete_inode(struct inode *);
extern struct dentry *rkfs_inode_lookup(struct inode *parent_inode, struct dentry *dentry, 
										struct nameidata *);
extern int rkfs_inode_create(struct inode *, struct dentry *, int mode,
							 struct nameidata *);
extern void rkfs_inode_truncate (struct inode * inode);
extern int rkfs_inode_unlink (struct inode *,struct dentry *);
extern int rkfs_inode_rename(struct inode * , struct dentry *, struct inode * , struct dentry *);


/** @} */


/** @name declaration of functions defines in fs_file.c
 *  @see fs_file.c
 * 
 * @{
 */
extern int rkfs_file_open (struct inode *, struct file *);
extern int rkfs_file_readdir(struct file *file, void *dirent, filldir_t filldir);
extern int rkfs_file_release (struct inode *, struct file *);
extern ssize_t rkfs_file_read(struct file *, char __user *, size_t, loff_t *);
extern ssize_t rkfs_file_write(struct file *, const char __user *, size_t, loff_t *);
extern loff_t rkfs_file_llseek (struct file * , loff_t , int );
/** @} */


/** @name declaration of functions defines in fs_dir.c
 *  @see fs_dir.c
 * 
 * @{
 */
extern int rkfs_mkdir(struct inode *, struct dentry *, int);
extern int rkfs_rmdir(struct inode *, struct dentry *);
/** @} */


/** @name declaration of functions defines in fs_lib.c
 *  @see fs_lib.c
 * 
 * @{
 */
int rkfs_next_data_item(char **, char *);
int rkfs_atoi(char*);
/** @} */


/** @name declaration of structures used by VFS
 * 
 * @{
 */
extern  struct super_operations rkfs_sops;
extern  struct inode_operations rkfs_iops;
extern  struct file_operations rkfs_fops;
extern  struct file_operations rkfs_dops;
extern  struct file_system_type rkfs;
extern  struct address_space_operations rkfs_aops;
/** @} */


/** @name different variables use by MySQLFS
 * 
 * @{
 */
int query(const char* command, int command_lenght);
void rkfs_deamon_test(void);
int rkfs_next_data_item(char **com_data_pos, char *ret_buf);
int rkfs_atoi(char* str);
extern int result;
extern struct_command ret_command;
extern char out_buf[OUT_BUF_LENGTH];
/** @} */


#endif


/**
 *
 * @}
 * 
 * @}
 * 
 */
