/**
 * @file parser.c
 * @brief Header file for implementation of functions and defines and structs, file fsd_lib.c
 * @author Michal Podzimek, Michal Ficek, Tomas Pop, MFF UK, 2007
 *
 * @addtogroup q_mysqlfs_parser Parser
 * @{
 * 
 * This file contents declaration of functions and definitions of constants and structures which are
 * used in mysql fs deamon code.
 * 
 */


#include "parser.h"


/* parser is one for userspace and one for kernel */
#ifdef __KERNEL__
 #include <linux/kernel.h>
 #include <linux/string.h>
#else
 #include <stdio.h>
 #include <string.h>
#endif


/**
 * @brief Translate command from text string (going thruechar device) to \ 
 * number representation
 * @param command_string Pointer to command string.
 * @return Command number 
 * @see fsd_defines.h
 * 
 */
inline int get_command_nr (const char * command_string){


/* commands sends from deamon to kernel */
#ifdef __KERNEL__	
	if (strcmp(command_string, CMD_OK) == 0 ){
		return CMD_OK_NR;	
	}
	
	if (strcmp(command_string, CMD_ER) == 0 ){
		return CMD_ER_NR;	
	}
	
/* commands sends from kernel to deamon */
#else 
	if (strcmp(command_string, TEST) == 0 ){ 	
		return TEST_NR;	
	}
	
	if (strcmp(command_string, CREATE) == 0 ){
		return CREATE_NR;	
	}
	
	if (strcmp(command_string, MKDIR) == 0 ){
		return MKDIR_NR;	
	}
	
	if (strcmp(command_string, CLEAR_FILE) == 0 ){
		return CLEAR_FILE_NR;	
	}
	
	if (strcmp(command_string, ADD_BLOCK) == 0 ){
		return ADD_BLOCK_NR;	
	}
	
	if (strcmp(command_string, REMOVE_FILE) == 0 ){
		return REMOVE_FILE_NR;	
	}
	
	if (strcmp(command_string, REMOVE_DIR) == 0 ){
		return REMOVE_DIR_NR;	
	}
	
	if (strcmp(command_string, GET_BLOCK) == 0 ){
		return GET_BLOCK_NR;	
	}
	
	if (strcmp(command_string, LOOKUP) == 0 ){
		return LOOKUP_NR;	
	}
	
	if (strcmp(command_string, READ_DIR) == 0 ){
		return READ_DIR_NR;	
	}
	
	if (strcmp(command_string, READ_INODE) == 0 ){
		return READ_INODE_NR;	
	}
	
	if (strcmp(command_string, REPLACE_BLOCK) == 0 ){
		return REPLACE_BLOCK_NR;	
	}
	
	if (strcmp(command_string, MOVE_FILE) == 0 ){
		return MOVE_FILE_NR;	
	}
	
	if (strcmp(command_string, RENAME_FILE) == 0 ){
		return RENAME_FILE_NR;	
	}

	if (strcmp(command_string, RENAME_DIR) == 0 ){
		return RENAME_DIR_NR;	
	}
	
	if (strcmp(command_string, TRUNCATE_FILE) == 0 ){
		return TRUNCATE_FILE_NR;	
	}
	

	
#endif	
	return UNKNOWN_COMMAND_NR;
}


/**
 * @brief Parse message to command structure.
 * 
 * This function parse message from fs kernel and put it to command structure.
 * Be sure, all array in the command structure are big enought 
 * 
 * @param message Pure message from fs kernel.
 * @param message_len Length of the message.
 * @param command Structure to be filled by this function.
 * @return OK on success or EPROTOKOL_VERSION_ERROR when protokol is not corect.
 * 
 */
int parse_message(char *message, int message_len, struct_command * command) {
	
	/* chars counter */
	int i = 0;
	int pos = 0;
	
	/* determine version of protocol */
	while ((pos < message_len) && ((*message) != ' ') && (i < MAX_VERSION_LENGTH) ) {
		command->version[i++] = *message;
		message++;
		pos++;
	}
	
	command->version[i] = '\x0';
	if ((*message) != '\x0') {
		message++;
		pos++;		
	}
	
	if (strcmp(command->version,FSD_VERSION) != 0){
		PDEBUG("fsd_parse_message - protokol version control failed\n");
		return EPROTOKOL_VERSION_ERROR;
	}
	
	i = 0;
	/* command */
	while  ((pos < message_len) && ((*message) != ' ') && (i < MAX_COMMAND_LENGTH)) {
		command->command[i++] = *message;
		message++; 
		pos++;
	}
	
	command->command[i] = '\x0';
	if ((*message) != '\x0') {
		message++;
		pos++;
	}
	
	command->command_nr = get_command_nr(command->command);
	
	i = 0;
	/* parameter */
	while ((pos < message_len) && ((*message) != ' ') && (i < MAX_PARAM_LENGTH)) {
		command->param[i++] = *message;
		message++;
		pos++;
		
	}
	
	command->param[i] = '\x0';
	if ((*message) != '\x0') {
		message++;
		pos++;
	}
	
	i = 0;
	/* data */
	while ((pos < message_len) && (i < (MAX_BLOCK_LENGTH + MAX_SIZE_LENGTH + 1))) {
		/* printf("pos=%d, i=%d, message[i]=%c\n", pos, i, *message); */
		command->data[i++] = *message;
		message++;
		pos++;
		
	}
	
	command->data[i] = '\x0'; /* added by Michal F. - zero is there for debugging messages */
	command->data_len = i;
	return OK;
}


/**
 *
 * @}
 * 
 */
